//                              -*- Mode: C++ -*-
//:ELNearClusters.C
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 27/5/97
// Last Update ........
//
//

#include "care/ELNearClusters.hh"
#include "care/ELNearHitCells.hh"
#include "care/EMClusterTree.hh"

#ifdef _OLD_NAMES_
#  include <algo.h>   
#else
#  include <algorithm>
#endif
#undef WRITE
/*
CLASS 
  ELNearClusters
  Get the list of clusters nearby a given cell
  and provide access to the list

*/

///////////////// ELNearClusters.C member function definitions ////////
//--
//  constructor
ELNearClusters::ELNearClusters() : 
  curr(0), numNeib(0) {}

//--
// constructor
ELNearClusters::ELNearClusters(ELNearHitCells & nearT) :
  curr(0), numNeib(0){
    for(nearT.begin(); nearT.more(); nearT.next())
       insert((*nearT)->getParent());
}
//--
//  destructor
ELNearClusters::~ELNearClusters(){}


//--
//--
//  Get the neighbours of a given cell
void ELNearClusters::of(const EMHitCell* hcell, const EMClusterTree & tree){
  //--
  // find the index of the cell pointed by the current cell
  of(hcell->cellIdx(), tree);
}

//--
//  Get neighbouring clusters of a given cell
void ELNearClusters::of(CellIdx_t cell, const EMClusterTree & tree){

  ELNearHitCells nearCells;
  nearCells.of(cell, tree); //  Get neibouring cells
  numNeib = 0;
  for(nearCells.begin(); nearCells.more(); nearCells.next())
    insert((*nearCells)->getParent());
  curr = 0;
}


//--
//  Insert the current cluster if it is not already present
void ELNearClusters::insert(EMCluster* clus){
  unsigned int i=0;
  for(; i<numNeib; i++)
    if( neibClusters[i]==clus) break;
  if( i==numNeib)
    neibClusters[numNeib++]=clus;
}

//--
//  Minimize the cluster list
//  This means take the cluster list and reduce it (following upstream
//  parent links) to a list of clusters with the same parent cluster.
//  Reduce the list eliminating the clusters equals and sort them in 
//  decreasing order in energy of the biggest signal.
void ELNearClusters::minimize(){
 // Start evaluating the minimum depth of the cluster stored
  Level_t minimumDepth = MAXDEPTH; // a very huge value;
  Level_t depth;
  unsigned int i;
  for( i=0; i<numNeib; i++){
    depth = neibClusters[i]->depthLevel();
    if( depth<minimumDepth) 
      minimumDepth = depth;
  }
  // now in minimumDepth there is the minimum depth of the clusters in the 
  // array
  // Now for each cluster in the array follow the parent link until a 
  // parent of depth equal to minimumDepth is reached
  for( i=0; i<numNeib; i++){
    while(neibClusters[i]->depthLevel() > minimumDepth){
      neibClusters[i] = neibClusters[i]->getParent();
      if(  neibClusters[i]==0){
	cout<<" Error: reached Top level in minimize"<<endl;
      } 
    }
  }
  // for each cluster found test if there are more than one occurrence:
  // if it is so put to 0 all occurreces after the first
  oneOccurrence();
  //
  //  reduce the array
  reduce();
  // now test if only a cluster has been found.
  // if not follow up for a level until level 1 is found or numberNNC = 1
  while(numNeib>1 && minimumDepth>FIRSTLEVEL){
    for(i=0; i<numNeib; i++)
      neibClusters[i] = neibClusters[i]->getParent();
    minimumDepth = neibClusters[0]->depthLevel();
    oneOccurrence();
    reduce();
  }
}

//--
//  Test if there are more than one occurrence of a given cluster
//  if it is so put to 0 all occurrencies after the first
void ELNearClusters::oneOccurrence(){
  for(unsigned int i=0; i<numNeib-1; i++)
    if( neibClusters[i]!=0){
      for(unsigned int j=i+1; j<numNeib; j++)
	if( neibClusters[i] == neibClusters[j] ){
	  neibClusters[j] = 0; // invalidate the pointer
	}
    }
}

//--
//  Get the number of occurrences of cluster
void ELNearClusters::reduce(){

  unsigned int j=0;  // count number of valid pointers
  for(unsigned int i=0; i<numNeib; i++){
    neibClusters[j] = neibClusters[i];
    if( neibClusters[i] != 0) j++;
  }
  numNeib = j;
}

//--
//  Sort decreasing in energy
void  ELNearClusters::sortClusters(){
  sort(&neibClusters[0], &neibClusters[numNeib], comp_cluster_pointers);
}








