//                              -*- Mode: C++ -*-
//: EMTrackOnEcal.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 9/5/97
// Last Update ........
//
#undef HIST

#ifdef HIST
#  include "ntuple.hh"
#endif
#include "care/EMTrackOnEcal.hh"
#include "care/care.hh"
#include "care/caredata.hh"
#include "care/EAClusterAnalyzer.hh"
#include "care/EDClusterModifier.hh"
#include "care/EMClusterTree.hh"
#include "care/ELLUTsManager.hh"
#include "care/ELCellGrid.hh"
#include "arte/ArteMessage.hh"
#include <math.h>

#ifdef _OLD_NAMES_
#  include <algo.h>   
#else
#  include <algorithm>
#endif

#ifdef PERFORMANCES
#include "arte/Chrono.hh"
extern Chrono*  extrap;
#endif

/*
CLASS
  EMEMTrackOnEcal
  Concrete derived class for the hierarchical cluster tree used in
  pattern recognition of ECAL. This class is a "leaf": it does not have
  children, it cannot receive Add, Move or Remove message.
  It is used to store the relation between an extrapolated track on the
  calorimeter and the EMHitCell in which the extrapolated point is. 
  

*/


///////////// Class EMTrackOnEcal member function definitions

//EMTrackOnEcal::

// constructor
EMTrackOnEcal::EMTrackOnEcal() : where(0) {}
// destructor
EMTrackOnEcal::~EMTrackOnEcal(){}

//--
// Return a flag to tell if this EMClusterMember is a leaf (if it is not, then
// it can be a clustermember list
bool EMTrackOnEcal::isALeaf() const { return true;}

//--
// Return always 0.0 to flag that this is a EMTrackOnEcal
//

Energy_t  EMTrackOnEcal::maxEnergy() const { 
    return 0.0; 
}


//--
// Analysis and traversing of the cluster tree is done by ClusterAnalyzers
// that are object with definite analyzing properties. 
// ClusterMembers MUST ignore these properties: in this way a loose coupling
// between ClusterMembers and ClusterAnalyzers is achieved and both
// can vary independently during program evolution.
// Examples of ClusterAnalyzers are those that should return
// total energy of a cluster, the Center of Gravity, the Number of Cells.. 
void EMTrackOnEcal::analyzeMe(EAClusterAnalyzer & analyzer) const {
  analyzer.doAnalysis(this);
}

void EMTrackOnEcal::modifyMe(EDClusterModifier & modifier){
  modifier.doModify(this);
}



//--
//  Dump for debugging
void EMTrackOnEcal::dump() const {

  int i=0;
  for(i=0; i<depthLevel(); i++)
    arteMsg<<"|";
  //  arteMsg.form("->Track %3d point (%6.1f,%6.1f,%8.1f) Mom = %8.5f ",
  //    ptOnEcal.track,
  //    ptOnEcal.xf,ptOnEcal.yf,ptOnEcal.zf,
  //    1./ptOnEcal.pf);
# ifdef CMATCH
  arteMsg<<"->Track "<<ptOnEcal->link()<<" point ("<< ptOnEcal->x()<<", "
	 <<ptOnEcal->y()<<", "<<ptOnEcal->z()<<") Mom = "<<1./ptOnEcal->pe();
# else
  arteMsg<<"->Track "<<ptOnEcal.track<<" point ("<< ptOnEcal.xf<<", "
	 <<ptOnEcal.yf<<", "<<ptOnEcal.zf<<") Mom = "<<1./ptOnEcal.pf;
# endif
  arteMsg.allp();
  for(i=0; i<depthLevel(); i++)
    arteMsg<<"|";
  if(where!=0){
    if( where->whichBank()!=0){
      arteMsg<<"        Mod "<<where->whichBank()->gcce;
    } else {
      arteMsg<<" Nowhere ";
    }
  } else {
    arteMsg<<" Nowhere ";
  }
# ifdef MCARLO
# ifdef CMATCH
  vector<ArtePointer<MTRA> > mtralis = ptOnEcal->mtraList();
  ArtePointer<MTRA> mctra(0);
  if( mtralis.begin()!= mtralis.end())
    mctra = *mtralis.begin();
# else
  if(ptOnEcal.mctrack!=0){
    //    arteMsg.form(" mc %d ",ptOnEcal.mctrack.index()); 
    arteMsg<<" mc "<<ptOnEcal.mctrack.index(); 
  }
  ArtePointer<MTRA> mctra = ptOnEcal.mctrack;
# endif
  if(mctra.index()!=0 && 
    mctra.index() < ArtePointer<MTRA>::nrow() ){
    ArtePointer<TOPP> topp = mctra->topp;
    float energy = mctra->mass*mctra->mass
      +  mctra->px*mctra->px
      +  mctra->py*mctra->py
      +  mctra->pz*mctra->pz;
    float e = sqrt(energy);
    char name[13];
    strncpy(name,topp->name,12);
    name[12] = 0;
    arteMsg<<name<<" energy:"<<e; 
  } else {
    arteMsg<<"  Ghost track ";
  }
  arteMsg.allp();
# endif
}
 
//--
//  Fill the internal data with the passed structure.
//  Normally the passed data refers to the last hit of the track. 5
void EMTrackOnEcal::fill(TrackInfo &lastHit){
  //
# ifdef CMATCH
  ptOnEcal = &lastHit;  // store data
# else
  ptOnEcal = lastHit;  // store data
# endif
  // set to zero reference pointers
  nextClusterMember = 0; 
  previousClusterMember =0;
  parentCluster = 0;
  // set depth level to zero
  depth = TOPLEVEL;
}

//--
//  Extrapolate the hit to the calorimeter. Return a flag to signal
//  if the hit is in a cluster or not (maybe outside ECAL or in a HOLE or
//  in a cell without signal).
EMHitCell* EMTrackOnEcal::extrapolate(EMClusterTree * tree){
  //
  // extrapolate charged tracks to the surface of the calorimeter
  //
  return extrapolate(ECalInfo.ecalp->z1, tree);
}

//--
//  Extrapolate the hit to the calorimeter. Return a flag to signal
//  if the hit is in a cluster or not (maybe outside ECAL or in a HOLE or
//  in a cell without signal).
EMHitCell* EMTrackOnEcal::extrapolate(Coord_t zCal, EMClusterTree * tree){
  //
  // extrapolate charged tracks to zCal using a straith line
  //
# ifdef CMATCH
  //  cout<<" extrapolation: "<<ptOnEcal->link()<<endl
  //    <<" Start: ("<<ptOnEcal->x()<<","<<ptOnEcal->y()<<","
  //    <<ptOnEcal->z()<<")"<<endl;
# ifdef PERFORMANCES
  extrap->start();
  int iErr = 0;
  ptOnEcal->prolongate (zCal, iErr, 1);
  extrap->stop();
# else
  int iErr = 0;
  ptOnEcal->prolongate (zCal, iErr, 1);
# endif
  // cout<<" Stop : ("<<ptOnEcal->x()<<","<<ptOnEcal->y()<<","
  //     <<ptOnEcal->z()<<")"<<endl;
  // Find if the point on the calorimeter has been clusterized
  where = tree->whichHitCell(ptOnEcal->x(),ptOnEcal->y());
# else
  ptOnEcal.xf += ptOnEcal.txf*(zCal-ptOnEcal.zf);
  ptOnEcal.yf += ptOnEcal.tyf*(zCal-ptOnEcal.zf);
  ptOnEcal.zf = zCal;
  
  // Find if the point on the calorimeter has been clusterized
  where = tree->whichHitCell(ptOnEcal.xf,ptOnEcal.yf);
# endif
  
  return where; // return true if it points 
                // to a (clusterized) EMHitCell
}

//--
//  return the cell in which the track has been extrapolate
//  Warning: You should call this member function only after that
//           the extrapolate() one has been called
//  Return a flag (value = -1) to signal if the extrapolation point 
//  is outside ECAL or in a HOLE.
CellIdx_t EMTrackOnEcal::whichCellIdx() const {
  //
  // the extrapolated point of the track is
  // in (ptOnEcal.xf, ptOnEcal.yf,  eCal.ecalp->z1)
  //
  ELLUTsManager LM;
  ELCellGrid::ErrorTypes ierr;
  // get the number of nearest neibouring cells of the current one
# ifdef CMATCH
  CellIdx_t ncell = LM.cellsGrid().cellIdx(ptOnEcal->x(), 
					   ptOnEcal->y(), ierr);  
# else
  CellIdx_t ncell = LM.cellsGrid().cellIdx(ptOnEcal.xf, ptOnEcal.yf, ierr);  
# endif
  if( ierr!= ELCellGrid::OK){
    ncell = 0;// to signal outside ECAL or in a HOLE.
  }
  return ncell;
} 

EMClusterMember* EMTrackOnEcal::appendTo(EMCluster* parent){
  EMCluster* pare = parent;
  while( pare!=0){
    pare->newTrack();
    pare = pare->getParent();
  }
  return EMClusterMember::appendTo(parent);
}


//--
//  evaluate the distance of the extrapolated track from center of cell
float EMTrackOnEcal::distancesq(const EMHitCell* hitCell) const{
#ifdef CMATCH
  float dx = hitCell->x() - ptOnEcal->x();
  float dy = hitCell->y() - ptOnEcal->y();
#else
  float dx = hitCell->x() - ptOnEcal.xf;
  float dy = hitCell->y() - ptOnEcal.yf;
#endif
  return dx*dx+dy*dy;
}

