//**********************************************************************/
// Created: 23.08.97,   Olya Igonkina   for   MARPLE package
//
// implementation of RcMsOfCesl class
//
//**********************************************************************/
#include <iostream.h>
#include <iomanip.h>
#include <math.h>

#include "marple/marple.hh"
#include <CLHEP/config/CLHEP.h>

#include "arte/ArteMessage.hh"

extern "C"{
 extern void FORTRAN_ROUTINE( rcmesm )(int *, double *);
}

RcMsOfGesl::RcMsOfGesl(const ArtePointer <GESL> gesl,
		       HepDouble ms,
		       HepDouble zcenter,
		       int insideMagnet )
  : fgesl(gesl), fms(ms) , fz_center(zcenter), finsideMagnet(insideMagnet)
{
  int indx = gesl - *ArteTable<GESL>::begin() +1 ;
  if( indx == 0 || indx > (int)ArteTable<GESL>::size() )
    {
      arteMsg << "MARPLE: RcMsOfGesl: superlayer " << indx 
	   << " is unknown";
      arteMsg.allp();
      return;
    }
  if( fabs(ms)<0.1e-5 )
      {
	double dms;
	FORTRAN_ROUTINE( rcmesm)( &indx, &dms);
	setms( dms);
      }
    if( fabs(zcenter)<0.1e-5  ) setzcenter( ( gesl->z1 + gesl->z2)/2 );  
    if( insideMagnet < 0  )   setMagnet( gesl->mfl);   
}


// check if the point rp is inside of superlayer
int RcMsOfGesl::check( Hep3Vector &rp ) const
{
  if( rp.z() < (gesl()->z1+MARPLE_NEGLIGIBLE) ||
      rp.z() > (gesl()->z2-MARPLE_NEGLIGIBLE) ) 
    { return 0; }

  int i = 0;
  double dz, a1, a2, b1, b2;
  switch ( gesl()->shp )
    {
    case 1 :
      // superlayer is a box 
      if( rp.x() > gesl()->x1 && rp.x() < gesl()->x2 &&
      	  rp.y() > gesl()->y1 && rp.y() < gesl()->y2 )
      	{
      	  i = 1;
      	}
      break;
    case 2 :
      //  superlayer is a X-trapezoid (ie. x  along z)
      dz = ( rp.z() - gesl()->z1 ) / ( gesl()->z2 - gesl()->z1 );
      a1 = gesl()->x1 + gesl()->dx * dz;
      a2 = gesl()->x2 - gesl()->dx * dz;
      if( rp.x() > a1 && rp.x() < a2 &&
      	  rp.y() > gesl()->y1 && rp.y() < gesl()->y2)
      	{
      	  i = 1;
      	}
      break;
    case 3 :
      //  superlayer is a trapezoid (ie. x and y varying along z)
      dz = ( rp.z() - gesl()->z1 ) / ( gesl()->z2 - gesl()->z1 );
      a1 = gesl()->x1 + gesl()->dx * dz;
      a2 = gesl()->x2 - gesl()->dx * dz;
      b1 = gesl()->y1 + gesl()->dy * dz;
      b2 = gesl()->y2 - gesl()->dy * dz;
      if( rp.x() > a1 && rp.x() < a2 &&
      	  rp.y() > b1 && rp.y() < b2)
      	{
      	  i = 1;
      	}
      break;
    case 5 :
      //  superlayer is a tube 
      a1 = gesl()->x2 - gesl()->x1;
      a2 = gesl()->x2 + gesl()->x1;
      b1 = gesl()->y2 - gesl()->y1;
      b2 = gesl()->y2 + gesl()->y1;
      if( rp.x() > a1 && rp.x() < a2 &&
      	  rp.y() > b1 && rp.y() < b2)
      	{
      	  i = 1;
      	}
      break;
    case 9 :
      // superlayer is spherical shell
      a1 = gesl()->x2 - gesl()->x1;
      a2 = gesl()->x2 + gesl()->x1;
      b1 = gesl()->y2 - gesl()->y1;
      b2 = gesl()->y2 + gesl()->y1;
      if( rp.x() > a1 && rp.x() < a2 &&
      	  rp.y() > b1 && rp.y() < b2)
      	{
      	  i = 1;
      	}
      break;
    default :
      arteMsg << "MARPLE: the unknown shape of GESL" << gesl();
      arteMsg.allp();
      break;
    }
  return i;
}



// printout average Multiple scattering for GESL table

ostream & operator << (ostream & os, const RcMsOfGesl &rms)
{
  os <<  " Multiple Scattering for GESL: "<< setw(2) << rms.gesl();
  os << " MS = " << setw(11)<< rms.ms() ;
  os << " Z_GESL = " << setw(6) << rms.zcenter() <<endl;
  return os ;
}

