//**********************************************************************/
// Created: 23.08.97,   Olya Igonkina   for   MARPLE package
//
// implementation of RcPoint class
//
//**********************************************************************/
#include <iomanip.h>
#include <algo.h>
#include <math.h>

#include <CLHEP/Matrix/Vector.h>

#include "arte/arte.hh"
#include "arte/ArteMessage.hh"

#ifdef MARPLEPERFORMANCES
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
Chrono * chRcPointProlongate = 0;
Chrono * chRcPointLineSwim   = 0;
Chrono * chRcPointSwimRunge  = 0;
Chrono * chRcPointRkmain     = 0;
Chrono * chRcPointEstimate   = 0;
#endif

#include "marple/marple.hh"
extern vector <RcMsOfGesl> msGeslList;

int RcPoint::dummy = 0;

// Destructor
RcPoint::~RcPoint() 
{}

RcPoint::RcPoint(const Hep3Vector &rp,
		 HepDouble tx ,
		 HepDouble ty ,
		 HepDouble pe ,
		 int mark)
  : Hep3Vector(rp), ftx(tx), fty(ty), fpe(pe), fcov5(5,1), fmark(mark)
{
  //  HepSymMatrix m5(5,1);
  fcov5 = fcov5 * MARPLE_INFINITY;
}  

RcPoint::RcPoint(const Hep3Vector &rp,
		 HepDouble tx ,
		 HepDouble ty ,
		 HepDouble pe ,
		 const HepSymMatrix &cov,
		 int mark)
  : Hep3Vector(rp), ftx(tx), fty(ty), fpe(pe), fcov5(5,0), fmark(mark)
{
  setcov(cov);
}  

RcPoint::RcPoint(const Hep3Vector &rp,
		 HepDouble tx ,
		 HepDouble ty ,
		 const HepSymMatrix &cov,
		 int mark)
  : Hep3Vector(rp), ftx(tx), fty(ty), fpe(0), fcov5(5,0), fmark(mark)
{
  setcov(cov);
  fcov5(1,5) = fcov5(2,5) = fcov5(3,5) = fcov5(4,5) = 0.;
  fcov5(5,5) = MARPLE_INFINITY;
}  

RcPoint::RcPoint( const Hep3Vector &rp,
		  HepDouble tx ,
		  HepDouble ty ,
		  HepDouble pe ,
		  const double cov_array[15],
		 int mark)
  : Hep3Vector(rp), ftx(tx), fty(ty), fpe(pe), fcov5(5,0), fmark(mark)
{
  setcov(cov_array);
}

RcPoint::RcPoint( const Hep3Vector &rp,
		  HepDouble tx ,
		  HepDouble ty ,
		  HepDouble pe ,
		  const float cov_array[15],
		  int mark)
  : Hep3Vector(rp), ftx(tx), fty(ty), fpe(pe), fcov5(5,0), fmark(mark)
{
  setcov(cov_array);
}

RcPoint::RcPoint( const Hep3Vector &rp,
		  HepDouble tx ,
		  HepDouble ty ,
		  const double cov_array[15],
		  int mark)
  : Hep3Vector(rp), ftx(tx), fty(ty), fpe(0.), fcov5(5,0), fmark(mark)
{
  setcov(cov_array);
  fcov5(1,5) = fcov5(2,5) = fcov5(3,5) = fcov5(4,5) = 0.;
  fcov5(5,5) = MARPLE_INFINITY;
}

RcPoint::RcPoint( const Hep3Vector &rp,
		  HepDouble tx ,
		  HepDouble ty ,
		  const float cov_array[15],
		  int mark)
  : Hep3Vector(rp), ftx(tx), fty(ty), fpe(0.), fcov5(5,0), fmark(mark)
{
  setcov(cov_array);
  fcov5(1,5) = fcov5(2,5) = fcov5(3,5) = fcov5(4,5) = 0.;
  fcov5(5,5) = MARPLE_INFINITY;
}

//-----------------------------------------------------------------------
//equal 1 if diagonal elements of cov() are not equal '0', otherwise 0

int RcPoint::covMatrixOk() const
{
  for( int i=1; i<= 5 ;++i )
    {
      if( cov()(i,i) == 0 ) return 0;
    }
  return 1;
}

//set diagonal elements of cov() to MARPLE_INFINITY instead of '0' 
int RcPoint::checkAndCorrectCov()
{
  int correct = 0;
  for( int i=1; i<= 5 ; ++i )
    {
      if( fabs(cov()(i,i)) < MARPLE_NEGLIGIBLE || 
	  cov()(i,i) > MARPLE_INFINITY ) 
	{
	  setcov( i,i, MARPLE_INFINITY) ;
	  correct = 1;
	}else if( cov()(i,i) < 0 ){
          static int idMesg = 2736542;
          arteMsg << "MARPLE warning: diagonal element of matrix is negative, "
                  << "take absolute value"<< endl;        
          arteMsg.allp(idMesg,10);
	  setcov( i,i, fabs(cov()(i,i)) ) ;
	  correct = 1;
    	  
	}

    }
  return correct;
}

//-----------------------------------------------------------------------
// prolongate rp point to the z coordinate
// depends on the list of multiple scattering
void RcPoint::prolongate( HepDouble z_next )
{  
  int iErr=0;
  prolongate( z_next, iErr);
}

void RcPoint::prolongate( HepDouble z_next, 
			  int & iErr, 
			  int linear, 
			  int imsGesl)
{  

# ifdef MARPLEPERFORMANCES
  if(chRcPointProlongate == 0){
    ChronoManager CM;
    chRcPointProlongate = CM.createChrono("prolongate ");
  }
  chRcPointProlongate->start();
# endif

  iErr = 0;
  if(msGeslList.empty() )
    { 
      arteMsg << "MARPLE warning: msGeslList is empty "<< endl
	   << "call marpleinit() for correct MARPLE initialization";
      arteMsg.allp();
    }
  
  // make correct numeration
  unsigned int iumsGesl;
  if(  imsGesl < 0  )
    { 
      iumsGesl = 0; 
    }else{
      if( imsGesl >= (int)msGeslList.size()  ) 
	iumsGesl = msGeslList.size() - 1; 
      else
	iumsGesl = imsGesl;	
    }
  // step I find the closest GESL plane
  int istep = 0;
  if( z() < msGeslList[iumsGesl].zcenter() )
    {
      for( istep=iumsGesl;  
	   istep>=0 && z() < msGeslList[istep].zcenter(); --istep )
	{}
      if( z() < z_next ) istep++; 
    } else {
      for( istep=iumsGesl;  z() > msGeslList[istep].zcenter() &&
	     istep < (int)msGeslList.size() ; ++istep )
	{}
      if( z() > z_next ) istep--;
    }
  if( istep < 0 ) 
    { 
      lineswim ( z_next); 
# ifdef MARPLEPERFORMANCES
      chRcPointProlongate->stop();
# endif
      return;
    }

  // step II prolongate till z_next
  // swim RcPoint to the middle of the next GESL
  // then add multiple scattering
  HepDouble angle_correction = 1 + tx()*tx() + ty()*ty(); 
  angle_correction = sqrt( angle_correction);

  int insideMagnet = msGeslList[istep].insideMagnet();
  
  // prolongate forward direction ( along z axis)
  if( z() < z_next )
    {
      // check if the start point is inside the medium
      HepDouble z_cor = correct_zback( z(), 
				       msGeslList[istep].gesl() );
      int doCorrection = 0;
      HepDouble tmpMs = 0;
      if( z_cor > MARPLE_NEGLIGIBLE )
	{
	  //correct the overestimation of the medium
	  doCorrection = 1;
	  tmpMs =  -msGeslList[istep].ms() * angle_correction *z_cor ;
	}else{
	  if( istep > 0 )
	    {
	      z_cor = correct_zforward( z(), 
					msGeslList[istep-1].gesl() );
	      if( z_cor > 0 ) 
		{
		  //correct underestimation of medium
		  doCorrection = 1;
		  tmpMs = msGeslList[istep].ms() * angle_correction *z_cor;
		}
	    }
	}
      // second step
      for( ;  msGeslList[istep].zcenter() < z_next && 
	      istep < (int)msGeslList.size();     ++istep)
	{
	  // prolongate even points with pe()=0 
	  //             to have correct matrix in magnet
	  if( (insideMagnet || msGeslList[istep].insideMagnet() ) 
	      && !linear )
	  {
	    swimRunge( msGeslList[istep].zcenter(), iErr);
	    if( iErr == 1 )
	      {
# ifdef MARPLEPERFORMANCES
		chRcPointProlongate->stop();
# endif
		return;
	      }
	  }else{
	      lineswim(  msGeslList[istep].zcenter());
	  }
	  insideMagnet = msGeslList[istep].insideMagnet();
	  if( msGeslList[istep].check(*this) )
	    { 
	      addMs( msGeslList[istep].ms() * angle_correction );
	      if( doCorrection )
		{  
		  addMs( tmpMs);
		  doCorrection = 0 ;
		}
	    }
	}
      // last step
      if( z() < z_next )
	{ 
	  if(  insideMagnet && !linear )
	    {
	      swimRunge( z_next, iErr );
	      if( iErr == 1 )
		{
# ifdef MARPLEPERFORMANCES
		  chRcPointProlongate->stop();
# endif
		  return;
		}
	    }else{
	      lineswim( z_next );
	    }
	}
      // check if the end point is inside the medium
      if( istep > 0 )
	{
	  z_cor = correct_zforward( z_next, 
				    msGeslList[istep-1].gesl() );
	}else{
	  z_cor = 0;
	}
      if( z_cor > 0 ) 
	{
	  //correct overestimation of medium
	  if(  msGeslList[istep-1].check(*this) )
	    { 
	      addMs( -msGeslList[istep-1].ms() * angle_correction *z_cor);
	    }
	}else{
	  if( (istep) < (int)msGeslList.size() )
	    {
	      z_cor = correct_zback( z_next, msGeslList[istep].gesl() );
	      if( z_cor > 0 )
		{
		  //correct the underestimation of the medium
		  if( msGeslList[istep].check(*this) )
		    { 
		      addMs( msGeslList[istep].ms() * 
			     angle_correction *z_cor);
		    }
		}
	    }
	}
    //
    // or prolongate back direction 
    //
    } else {
      // check if the start point is inside the medium
      HepDouble z_cor = correct_zforward( z(), 
					  msGeslList[istep].gesl() );
      int doCorrection = 0;
      HepDouble tmpMs = 0;
      if( z_cor > 0 ) 
	{
	  //correct the overestimation of the medium
	  doCorrection = 1;
	  tmpMs = -msGeslList[istep].ms() * angle_correction *z_cor ;
	}else{
	  if( (istep+1) <(int) msGeslList.size() )
	    {
	      z_cor = correct_zback( z(), 
				     msGeslList[istep+1].gesl() );
	      if( z_cor > 0 )
		{
		  //correct the underestimation of the medium
		  doCorrection = 1;
		  tmpMs = msGeslList[istep].ms() * angle_correction *z_cor;
		}
	    }
	}
      // second step
      for( ;  msGeslList[istep].zcenter() > z_next && istep>=0; istep--)
	{
	  if( (insideMagnet || msGeslList[istep].insideMagnet()) 
	      && !linear )
	    {
	      swimRunge( msGeslList[istep].zcenter(), iErr);
	      if( iErr == 1 )
		{
# ifdef MARPLEPERFORMANCES
		  chRcPointProlongate->stop();
# endif
		  return;
		}
	    }else{
	      lineswim(  msGeslList[istep].zcenter());
	    }
	  insideMagnet = msGeslList[istep].insideMagnet();
	  if( msGeslList[istep].check(*this) )
	    { 
	      addMs( msGeslList[istep].ms() * angle_correction ); 
	      if( doCorrection )
		{  
		  addMs( tmpMs);
		  doCorrection = 0 ;
		}
	    }
	}
      // last step
      if( z() > z_next )
	{ 
	  if( insideMagnet && !linear )
	    {
	      swimRunge(z_next, iErr );
	      if( iErr == 1 )
		{
# ifdef MARPLEPERFORMANCES
		  chRcPointProlongate->stop();
# endif
		  return;
		}
	    }else{
	      lineswim(  z_next);
	    }
	}
      //take into account the medium
      if( istep >=0 )
	{
	  z_cor = correct_zback( z_next, msGeslList[istep].gesl() );
	}else{
	  z_cor = 0;
	}
      if( z_cor > 0 )
	{
	  //correct the overestimation of the medium
	  if( istep>0 && msGeslList[istep-1].check(*this) )
	    { 
	      addMs( -msGeslList[istep].ms() * angle_correction *z_cor);
	    }
	}else{
	  if( istep > 0 )
	    {
	      z_cor = correct_zforward( z(), 
					msGeslList[istep-1].gesl() );
	      if( z_cor > 0 && msGeslList[istep-1].check(*this) ) 
		{
		  //correct underestimation of medium
		  addMs( msGeslList[istep].ms() 
			 * angle_correction *z_cor);
		}
	    }
	}
    }

# ifdef MARPLEPERFORMANCES
  chRcPointProlongate->stop();
# endif

}

//------------------------------------------------------------------
/*

void RcPoint::prolongate2( HepDouble z_next, 
			   int & iErr, 
			   int linear ) 
{  

# ifdef MARPLEPERFORMANCES
  if(chRcPointProlongate == 0){
    ChronoManager CM;
    chRcPointProlongate = CM.createChrono("prolongate ");
  }
  chRcPointProlongate->start();
# endif

  iErr = 0;
  if(msGeslList.empty() )
    { 
      arteMsg << "MARPLE warning: msGeslList is empty "<< endl
	   << "call marpleinit() for correct MARPLE initialization";
      arteMsg.allp();
    }
  
  // step I find the closest GESL plane
  int istep = 0;
  if( z() < z_next      ) // 
    {
      for( istep = msGeslList.back().zcenter() ;  
	   z() < msGeslList[istep].zcenter() && istep>=0 ; --istep )
	{}
      if( z() < z_next ) istep++; 
    } else {
      for( istep=0;  z() > msGeslList[istep].zcenter() &&
	     istep < (int)msGeslList.size() ; ++istep )
	{}
      if( z() > z_next ) istep--;
    }
  if( istep < 0 || istep > (int)msGeslList.size() ) 
    { 
      lineswim ( z_next); 
# ifdef MARPLEPERFORMANCES
      chRcPointProlongate->stop();
# endif
      return;
    }

  //-----------------------------------------------
  // step II prolongate till z_next
  // swim RcPoint to the middle of the next GESL
  // then add multiple scattering

  HepDouble angle_correction = 1 + tx()*tx() + ty()*ty(); 
  angle_correction = sqrt( angle_correction);

  int insideMagnet = msGeslList[istep].insideMagnet();
  
  // prolongate forward direction ( along z axis)
  if( z() < z_next )
    {
      if( msGeslList[istep].check(z() ) // point inside medium
	  {
	    HepDouble z_cor = correct_zback( z(), 
					     msGeslList[istep].gesl() );
	    if( z_cor > MARPLE_NEGLIGIBLE )
	      {
		//correct the overestimation of the medium
		doCorrection = 1;
		tmpMs =  -msGeslList[istep].ms() * angle_correction *z_cor ;
	      }else{
	  if( istep > 0 )
	    {
	      z_cor = correct_zforward( z(), 
					msGeslList[istep-1].gesl() );
	      if( z_cor > 0 ) 
		{
		  //correct underestimation of medium
		  doCorrection = 1;
		  tmpMs = msGeslList[istep].ms() * angle_correction *z_cor;
		}
	    }
	}
      // second step
      for( ;  msGeslList[istep].zcenter() < z_next && 
	      istep < (int)msGeslList.size();     ++istep)
	{
	  // prolongate even points with pe()=0 
	  //             to have correct matrix in magnet
	  if( (insideMagnet || msGeslList[istep].insideMagnet() ) 
	      && !linear )
	  {
	    swimRunge( msGeslList[istep].zcenter(), iErr);
	    if( iErr == 1 )
	      {
# ifdef MARPLEPERFORMANCES
		chRcPointProlongate->stop();
# endif
		return;
	      }
	  }else{
	      lineswim(  msGeslList[istep].zcenter());
	  }
	  insideMagnet = msGeslList[istep].insideMagnet();
	  if( msGeslList[istep].check(*this) )
	    { 
	      addMs( msGeslList[istep].ms() * angle_correction );
	      if( doCorrection )
		{  
		  addMs( tmpMs);
		  doCorrection = 0 ;
		}
	    }
	}
      // last step
      if( z() < z_next )
	{ 
	  if(  insideMagnet && !linear )
	    {
	      swimRunge( z_next, iErr );
	      if( iErr == 1 )
		{
# ifdef MARPLEPERFORMANCES
		  chRcPointProlongate->stop();
# endif
		  return;
		}
	    }else{
	      lineswim( z_next );
	    }
	}
      // check if the end point is inside the medium
      if( istep > 0 )
	{
	  z_cor = correct_zforward( z_next, 
				    msGeslList[istep-1].gesl() );
	}else{
	  z_cor = 0;
	}
      if( z_cor > 0 ) 
	{
	  //correct overestimation of medium
	  if(  msGeslList[istep-1].check(*this) )
	    { 
	      addMs( -msGeslList[istep-1].ms() * angle_correction *z_cor);
	    }
	}else{
	  if( (istep) < (int)msGeslList.size() )
	    {
	      z_cor = correct_zback( z_next, msGeslList[istep].gesl() );
	      if( z_cor > 0 )
		{
		  //correct the underestimation of the medium
		  if( msGeslList[istep].check(*this) )
		    { 
		      addMs( msGeslList[istep].ms() * 
			     angle_correction *z_cor);
		    }
		}
	    }
	}
    //
    // or prolongate back direction 
    //
    } else {
      // check if the start point is inside the medium
      HepDouble z_cor = correct_zforward( z(), 
					  msGeslList[istep].gesl() );
      int doCorrection = 0;
      HepDouble tmpMs = 0;
      if( z_cor > 0 ) 
	{
	  //correct the overestimation of the medium
	  doCorrection = 1;
	  tmpMs = -msGeslList[istep].ms() * angle_correction *z_cor ;
	}else{
	  if( (istep+1) <(int) msGeslList.size() )
	    {
	      z_cor = correct_zback( z(), 
				     msGeslList[istep+1].gesl() );
	      if( z_cor > 0 )
		{
		  //correct the underestimation of the medium
		  doCorrection = 1;
		  tmpMs = msGeslList[istep].ms() * angle_correction *z_cor;
		}
	    }
	}
      // second step
      for( ;  msGeslList[istep].zcenter() > z_next && istep>=0; istep--)
	{
	  if( (insideMagnet || msGeslList[istep].insideMagnet()) 
	      && !linear )
	    {
	      swimRunge( msGeslList[istep].zcenter(), iErr);
	      if( iErr == 1 )
		{
# ifdef MARPLEPERFORMANCES
		  chRcPointProlongate->stop();
# endif
		  return;
		}
	    }else{
	      lineswim(  msGeslList[istep].zcenter());
	    }
	  insideMagnet = msGeslList[istep].insideMagnet();
	  if( msGeslList[istep].check(*this) )
	    { 
	      addMs( msGeslList[istep].ms() * angle_correction ); 
	      if( doCorrection )
		{  
		  addMs( tmpMs);
		  doCorrection = 0 ;
		}
	    }
	}
      // last step
      if( z() > z_next )
	{ 
	  if( insideMagnet && !linear )
	    {
	      swimRunge(z_next, iErr );
	      if( iErr == 1 )
		{
# ifdef MARPLEPERFORMANCES
		  chRcPointProlongate->stop();
# endif
		  return;
		}
	    }else{
	      lineswim(  z_next);
	    }
	}
      //take into account the medium
      if( istep >=0 )
	{
	  z_cor = correct_zback( z_next, msGeslList[istep].gesl() );
	}else{
	  z_cor = 0;
	}
      if( z_cor > 0 )
	{
	  //correct the overestimation of the medium
	  if( istep>0 && msGeslList[istep-1].check(*this) )
	    { 
	      addMs( -msGeslList[istep].ms() * angle_correction *z_cor);
	    }
	}else{
	  if( istep > 0 )
	    {
	      z_cor = correct_zforward( z(), 
					msGeslList[istep-1].gesl() );
	      if( z_cor > 0 && msGeslList[istep-1].check(*this) ) 
		{
		  //correct underestimation of medium
		  addMs( msGeslList[istep].ms() 
			 * angle_correction *z_cor);
		}
	    }
	}
    }

# ifdef MARPLEPERFORMANCES
  chRcPointProlongate->stop();
# endif

}
*/

//-----------------------------------------------------------------------
//swim  the RcPoint to the z_next plane without multiple scattering

// assume that magnet parameters are :
// z_magnet = 450. , r_magnet = 140., Hmag =8.

void RcPoint::fastswim( HepDouble z_next, int & iErr )
{
  static HepDouble ZPOSMG = 450;
  static HepDouble RMAG   = 160; // make region of interest for a swim a little bit bigger 
  //static HepDouble RMAG   = 140;
  static HepDouble HMAG   = 8; 

  static HepDouble ZMAGBIG = ZPOSMG + RMAG;
  static HepDouble ZMAGSMALL = ZPOSMG - RMAG;
  static HepDouble RMAG2   = RMAG*RMAG;

  iErr = 0;
  //if( pe() )
  //{ lineswim( z_next);    return; }

  // all time outside magnet 
  if( z_next > ZMAGBIG && z() > ZMAGBIG ) 
    { lineswim( z_next );   return ; }
  else   if( z_next < ZMAGSMALL && z() < ZMAGSMALL ) 
    { lineswim( z_next );   return ; }

  cout << " step 0 " << ends;

  //if( (z_next - z() ) * ( z_next - ZPOSMG - RMAG ) < 0 || 
  //  (z_next - z() ) * ( z_next - ZPOSMG + RMAG ) < 0  )
  //{ lineswim( z_next );   return ; }

  // step I - swiming outside the magnet till its border
  if( (x()*x() + ( z()-ZPOSMG)*(z()-ZPOSMG) ) > RMAG2 )
    {
      HepDouble AA = tx()*tx() + 1;
      HepDouble B1 = x() - tx()*z();
      HepDouble BB = tx()*B1 - ZPOSMG; 

      HepDouble DD = BB*BB - AA*(B1*B1 + ZPOSMG*ZPOSMG - RMAG2);
      if( DD < MARPLE_NEGLIGIBLE ) // no cross with magnetic field
	{  lineswim(z_next);  return;   }
      if( z() < ZPOSMG )
	{
	  HepDouble z_tmp = ( -BB - sqrt(DD) ) / AA ;
	  if( z_next < z_tmp )
	    { 
	      lineswim(z_next);
	      return;
	    }else{
	      lineswim(z_tmp);
	    }
	}else{
	  HepDouble z_tmp = ( -BB + sqrt(DD) ) / AA ;
	  if( z_next > z_tmp )
	    { 
	      lineswim(z_next);
	      return;
	    }else{
	      lineswim(z_tmp);
	    }
	}
    }

  cout << " step 1 " << ends;
  // step II - swiming inside the magnet 

  HepDouble rho = 3336.1134 / HMAG * sqrt( 1+tx()*tx() ) /
                  sqrt( 1 + tx()*tx() + ty()*ty() ) ;
  // pe should be equal Q/p 
  if (fabs(pe()) > MARPLE_NEGLIGIBLE )
    {
      rho /= pe();
    }else{
      //arteMsg << "MARPLE : swiming through magnet for point with Q/p = 0";
      //arteMsg.allp();
      lineswim(z_next);
      return;
    }

  cout << " step 2 " << ends;
  HepDouble XR = x();
  if( fabs( tx()) > MARPLE_NEGLIGIBLE )
    {
      XR += -rho * tx()/ fabs(tx()) / sqrt( 1 + tx()*tx()) ;
    }
  HepDouble YR = tx()*( x()-XR ) + z();

  HepDouble B1 = (RMAG2 -ZPOSMG*ZPOSMG +XR*XR +YR*YR -rho*rho)/2;
  HepDouble AA = (ZPOSMG - YR)*(ZPOSMG - YR) + XR*XR;
  HepDouble BB = (ZPOSMG-YR)*B1 -XR*XR*ZPOSMG;
  
  HepDouble DD = BB*BB - AA*(B1*B1 -XR*XR*( RMAG2 -ZPOSMG*ZPOSMG));
  //the particle is assumed to staying in magnet
  if( DD < MARPLE_NEGLIGIBLE )
    {  swimRunge(z_next, iErr); return; }
  cout << " step 3 " << ends;
  if( z() < ZPOSMG )
    {
      HepDouble z_tmp = ( -BB + sqrt(DD) ) / AA ; // anothe side if magnet 
      if( z_next < z_tmp )
	{ 
	  swimRunge(z_next, iErr);
	  return;
	}else{
	  swimRunge(z_tmp, iErr);
	  if( iErr == 1 ) return;
	}
    }else{
      HepDouble z_tmp = ( -BB - sqrt(DD) ) / AA ;
      if( z_next > z_tmp )
	{ 
	  swimRunge(z_next, iErr);
	  return;
	}else{
	  swimRunge(z_tmp, iErr);
	  if( iErr == 1 ) return;
	}
    }

  cout << " step 4 " << ends;
  //step III - swiming outside the magnet
  lineswim( z_next);
  cout << " step 5 " << ends;
}


//-----------------------------------------------------------------------
// linear prolongation of the RcPoint parameters to the z_next plane.

void RcPoint::lineswim( HepDouble z_next)
{
  // transport parameters 
  HepDouble dz = z_next - z();
  if( fabs(dz) < 1.e-4) return;

# ifdef MARPLEPERFORMANCES
  if(chRcPointLineSwim == 0){
    ChronoManager CM;
    chRcPointLineSwim = CM.createChrono("lineswim ");
  }
  chRcPointLineSwim->start();
# endif

  Hep3Vector shift( dz*tx(), dz*ty(), dz);  // x = x + dz * tx
  Hep3Vector::operator += (shift);
  
  // transport matrix
  // assistance matrix for transport F

  HepDouble dz2 = dz*dz; 
  setcov(1,1, cov()(1,1)+cov()(3,3)*dz2 +2*cov()(1,3)*dz);
  setcov(1,2, cov()(1,2)+cov()(3,4)*dz2 + (cov()(2,3) + cov()(1,4))*dz);
  setcov(1,3, cov()(1,3)        +          cov()(3,3)*dz);
  setcov(1,4, cov()(1,4)        +          cov()(4,3)*dz);
  setcov(1,5, cov()(1,5)        +          cov()(5,3)*dz);

  setcov(2,2, cov()(2,2)+cov()(4,4)*dz2 +2*cov()(2,4)*dz);
  setcov(2,3, cov()(2,3)        +          cov()(3,4)*dz);
  setcov(2,4, cov()(2,4)        +          cov()(4,4)*dz);
  setcov(2,5, cov()(2,5)        +          cov()(5,4)*dz);


  //  static HepMatrix cov_correction(5,5,1); 
  //  cov_correction[0][2] = dz;  // (1,3)
  //  cov_correction[1][3] = dz;  // (2,4)
  //  setcov( cov().similarity(cov_correction) );

# ifdef MARPLEPERFORMANCES
  chRcPointLineSwim->stop();
# endif

}

//-----------------------------------------------------------------------
extern "C" 
{ extern void FORTRAN_ROUTINE( rkmain )( int*,
					 double*,
					 double*,
					 double [6],
					 double [4],
					 double [6],
					 double [5][5],
					 int*);
}

void RcPoint::swimRunge( HepDouble z_next,   
			 int & iErr,
			 HepDouble FFstep )
{
  iErr = 0;
  double distanceCor = fabs( z_next - z() );
  if( distanceCor < 0.1e-5 ) return;
  
  double Fstep;
  if( FFstep ) 
    {
      Fstep = FFstep;
    }else{
      Fstep = distanceCor;
    }
  if( Fstep < 0.1e-5 ){   return;  }

# ifdef MARPLEPERFORMANCES
  if(chRcPointSwimRunge == 0){
    ChronoManager CM;
    chRcPointSwimRunge = CM.createChrono("swimRunge ");
    chRcPointRkmain = CM.createChrono("rkmain ");
  }
  chRcPointSwimRunge->start();
  chRcPointRkmain->start();
# endif

  //be sure to use Q/p as pe() value !!!
  double Pin[6]  = { z(), x(), y(), tx(), ty(), pe() };
  if( distanceCor < 1 ) distanceCor = 1;
  double Prec[4] = { 0.001/distanceCor, 0.001/distanceCor, 
		     0.0001/distanceCor, 0.0001/distanceCor};
  static double Pout[6] ;
  static double RKD[5][5];

  static int Irkmain = 2; // default value - fast calculation of matrix
  FORTRAN_ROUTINE( rkmain )(&Irkmain,
                            &z_next, &Fstep, Pin, Prec, 
                            Pout, RKD, &iErr );

# ifdef MARPLEPERFORMANCES
    chRcPointRkmain->stop();
# endif

  switch (iErr )
    {
    case 1 :
      arteMsg << " MARPLE : track curls during Runge-Kutta tracking" ;
      arteMsg.allp();
      break;
    case 2 :
      arteMsg << " MARPLE : zero FSTEP was used";
      arteMsg.allp();
      break;
    case 3 :
      arteMsg << " MARPLE :rkstep call limit exceeded";
      arteMsg.allp();
      break;
    }
  setX(  Pout[1]);
  setY(  Pout[2]);
  setZ(  z_next );
  settx( Pout[3]);
  setty( Pout[4]);
  setpe( Pout[5]);



  static HepMatrix cov_correction(5,5);
  for(int i=0; i<5; ++i )
    for(int j=0; j<5; ++j )
      {
	// C++ style of array is opposite to fortran's one
	// when we use '[]' the numeration starts from 0
	cov_correction[i][j] = RKD[j][i];
      }    
  setcov( cov().similarity(cov_correction) );

  //  cout << " Similarity \n" << cov().similarity(cov_correction) << endl;
  

  

# ifdef MARPLEPERFORMANCES
    chRcPointSwimRunge->stop();
# endif


}

//-----------------------------------------------------------------------
// correct covariance matrix by the multiple scattering 

void RcPoint::lineMsCorr( unsigned int iGeslMin,
			  unsigned int iGeslMax, 
			  HepDouble zend,
			  int forward)
{

  /* There is a way to estimate multiple scattering income
     to the covariant matrix during prolongation on the long distance dz  
     if prolongation is linear and the planes on the way are thin 

     Correction  = Sum_i ( ms_i * |  dz*dz    0     dz   0   0  |  )
                    ^             |    0    dz*dz   0    dz  0  |_i
		    |	          |    dz     0     1     0  0  |
		over planes       |    0      dz    0     1  0  |
		                  |    0      0     0     0  1  |


     */

  HepDouble dz, corr;

  HepDouble angle_correction = 1 + tx()*tx() + ty()*ty(); 
  angle_correction = sqrt( angle_correction);
  
  for( unsigned int igesl = iGeslMin; igesl < iGeslMax; ++igesl )
    {    
      // to prolongate forward the center of igesl should be < zend
      dz = zend - msGeslList[igesl].zcenter();
      if( forward * dz < 0 )continue;
      corr = msGeslList[igesl].ms() * angle_correction;
      setcov(3,3, cov()(3,3)+ corr ); // ~ ms
      setcov(4,4, cov()(4,4)+ corr ) ;
      corr *= dz;
      setcov(1,3, cov()(1,3)+ corr );  // ~ ms*dz
      setcov(2,4, cov()(2,4)+ corr );
      corr *= dz;
      setcov(1,1, cov()(1,1)+ corr );  // ~ ms*dz*dz
      setcov(2,2, cov()(2,2)+ corr );
    }

}


void RcPoint::addMs( const HepDouble ms )
{
  // use pe() = Q/p !!! if it is zero, set 1 !
  if( fabs(pe()) > MARPLE_NEGLIGIBLE )
    { 
      HepDouble correction = ms*pe()*pe() ;
      setcov( 3, 3, cov()(3,3) + correction );
      setcov( 4, 4, cov()(4,4) + correction );
    } else {
      HepDouble correction = ms;
      setcov( 3, 3,  cov()(3,3) + correction );
      setcov( 4, 4,  cov()(4,4) + correction );
    }


}

//-----------------------------------------------------------------------
// Kalman filter to update information with measured value HITB ( CHANGE RcPoint), return $\Delta \chi^2$ of procedure. 
HepDouble RcPoint::filter( ArtePointer< HITB> hitb, int & iErr)
{
  iErr = 0;
  prolongate( HepDouble(hitb->z), iErr );
 
  if( iErr ) 
    {
      arteMsg << "MARPLE: prolongation is impossible, filter aborted";
      arteMsg.allp();
      return 0 ; 
    }
  HepDouble sig = hitb->r; // the accuracy of SVD measurement
  HepDouble sig2 = sig*sig ;
  ArtePointer<GEDE> gede(hitb->gede);
  HepDouble co =  gede->cos*gede->rot[0] - gede->sin*gede->rot[1];
  HepDouble si =  gede->sin*gede->rot[0] + gede->cos*gede->rot[1];
  HepVector Hvec(5,0);
  HepDouble residual, hCh;
  HepDouble mU, Hp, cosThetaU, sicoThetaU;
  
  int cmp = gede->cmp;
  switch ( cmp ) 
    {
    case 1 :
    case 3 :
      residual = hitb->x - co* x() + si * y();
      Hvec(1) = co;
      Hvec(2) = -si;
      
      hCh = cov()(1,1)*co*co + cov()(2,2)*si*si
	- 2*cov()(1,2)*co*si ;
      if( hCh <= 0 )
	{ 
	  int idMess = 345235;
	  arteMsg << "MARPLE: filter: matrix is not pos-defined!";
	  arteMsg.allp(idMess, 10);
	  hCh = fabs(hCh);
	  iErr = 10;
	}
      break;
    case 2 :      
      mU = tx()*co - ty()*si; // tan ( track angle) in uz-plane
      cosThetaU = 1 / sqrt( 1 + mU*mU ); 
      sicoThetaU = mU * cosThetaU * cosThetaU;
      
      Hp = ( x()*co - hitb->x  - y()*si ) * cosThetaU; //xwire = hitb_x/co
      Hvec(1) =  co * cosThetaU;
      Hvec(2) = -si * cosThetaU;
      Hvec(3) = -co   * sicoThetaU * Hp;
      Hvec(4) = -(-si)* sicoThetaU * Hp;
      
      residual = hitb->q - Hp;
      hCh = cov().similarity(Hvec);
      break;
      
    default :
      arteMsg << "MARPLE: filter: no rules to add HITB " << hitb 
	   << " with CMP = " << cmp << " to RcPoint exist" ;
      arteMsg.allp();
      iErr = 11;
      return 0;
      break;
    }
  

  HepVector Gain(5,0);
  Gain = cov()*Hvec;
  Gain /= sig2+hCh ;
  
  setX(  x()  + Gain(1)*residual );
  setY(  y()  + Gain(2)*residual );
  settx( tx() + Gain(3)*residual );
  setty( ty() + Gain(4)*residual );
  setpe( pe() + Gain(5)*residual );
  
  HepMatrix correctM(5,5,1);
  correctM = correctM - Gain*Hvec.T();
  correctM = correctM * cov() ;
  HepSymMatrix newcov(5,0);
  newcov.assign(correctM);
  setcov( newcov );

  // chi2 = [(1-Hvec.T()*Gain)*residual]**2 / [(1-Hvec.T()*Gain)*sig2]
  
  HepDouble deltaChi2 = Gain(1)*Hvec(1) + Gain(2)*Hvec(2) +
    Gain(3)*Hvec(3) + Gain(4)*Hvec(4) + Gain(5)*Hvec(5); 
  deltaChi2 = residual * residual * ( 1 - deltaChi2) / sig2;

  return deltaChi2;
}

//-----------------------------------------------------------------------
bool comp_hitb_z( const ArtePointer<HITB> hitb1, 
		  const ArtePointer<HITB> hitb2)
{  return hitb1->z < hitb2->z ;    } 

// Kalman refit, return $\chi^2$ of procedure ( CHANGE RcPoint). 
HepDouble RcPoint::refit(int & iErr)
{
  iErr = 0;
  HepDouble deltaChi2 = 0;

  vector<ArtePointer<HITB> > hitbL = hitbList();
  sort( hitbL.begin(), hitbL.end(), comp_hitb_z);
  HepDouble z_first = hitbL[0]->z;
  HepDouble z_end = hitbL[ hitbL.size()-1 ]->z;

  int iErr1 = 0;
  // refit started from closest end of HITB vector
  if( fabs( z() - z_first) < fabs( z() - z_end) )
    {  
      for( unsigned int iHit=0; iHit < hitbL.size(); ++iHit )
	{ 
	  deltaChi2 += filter( hitbL[iHit], iErr1 ); 
	  if( iErr1 ) iErr = iErr1;
	}
    }else{
      for( unsigned int iHit=hitbL.size(); iHit > 0; iHit-- )
	{ 
	  deltaChi2 += filter( hitbL[iHit-1], iErr1  ); 	
	  if( iErr1 ) iErr = iErr1;
	}
    }
  return deltaChi2;
}

//-----------------------------------------------------------------------
HepDouble RcPoint::refitForward(int & iErr)
{
  iErr = 0;
  HepDouble deltaChi2 = 0;
  vector<ArtePointer<HITB> > hitbL = hitbList();
  sort( hitbL.begin(), hitbL.end(), comp_hitb_z);

  int iErr1 = 0;
  // refit started from closest Z of HITB vector to bigger one
  unsigned int iHit=0;
  for(; iHit < hitbL.size() && (hitbL[iHit]->z) < z(); ++iHit )
    {}
  for( ; iHit < hitbL.size() ; ++iHit )
    {
      deltaChi2 += filter( hitbL[iHit], iErr1 );
      if(iErr1 ) iErr = iErr1; 
    }
   
  return deltaChi2;
}

HepDouble RcPoint::refitBack(int & iErr)
{
  iErr = 0;
  HepDouble deltaChi2 = 0;
  vector<ArtePointer<HITB> > hitbL = hitbList();
  sort( hitbL.begin(), hitbL.end(), comp_hitb_z);

  // refit started from closest Z of HITB vector to smaller one
  unsigned int iHit=hitbL.size();
  for(; iHit > 0 && (hitbL[iHit-1]->z > z()); iHit-- )
    {}
  int iErr1 = 0;
  for( ; iHit > 0; iHit-- )
    {  
      deltaChi2 += filter( hitbL[iHit-1], iErr1  );     
      if( iErr1 ) iErr = iErr1;
    }
  
  return deltaChi2;
}

//----------------------------------------------------------------------------

//estimate momentum using target information and correct it (CHANGE RcPoint)
// using Peter Weyers algorithm
void RcPoint::estimatePEfromTAR( HepDouble tarX, HepDouble tarY, 
				 HepDouble tarZ,
				 HepDouble covTarXX, HepDouble covTarYY, 
				 HepDouble covTarXY,
				 int nMaxIter, 
				 int & iErr,
				 int coordinates)
{
  estimatePARfromTAR( 5, tarX, tarY, tarZ, 
		      covTarXX, covTarYY, covTarXY,
		      nMaxIter, iErr, coordinates );
}

void RcPoint::estimatePARfromTAR( int iPar,
				  HepDouble tarX, HepDouble tarY, 
				  HepDouble tarZ,
				  HepDouble covTarXX, HepDouble covTarYY, 
				  HepDouble covTarXY,
				  int nMaxIter, 
				  int & iErr,
				  int coordinates)
{

# ifdef MARPLEPERFORMANCES
  if(chRcPointEstimate == 0){
    ChronoManager CM;
    chRcPointEstimate = CM.createChrono("estimatePar ");
  }
  chRcPointEstimate->start();
# endif



  HepDouble zOld = z();

  iErr = 0; // error code

  // temporary variables
  RcPoint tmpPoint;
  HepSymMatrix invCov(5,1);

  HepDouble deltaX = 0;
  HepDouble deltaY = 0;
  
  // inverted covariant matrix of target
  HepDouble  icovTarXX, icovTarYY, icovTarXY;
  if( covTarXY == 0 )
    {
      if( covTarXX ) icovTarXX = 1 / covTarXX; else icovTarXX = 0;
      if( covTarYY ) icovTarYY = 1 / covTarYY; else icovTarYY = 0;
      icovTarXY = 0;
    }else{
      HepDouble determinant = 1/(covTarXX*covTarYY-covTarXY*covTarXY);
      if( fabs(determinant) < MARPLE_NEGLIGIBLE ) determinant = 1;
      icovTarXY = -covTarXY / determinant;
      HepDouble tmpXX = covTarXX;
      icovTarXX = covTarYY/determinant;
      icovTarYY = tmpXX/determinant;
    }
  
  HepDouble parNew = 0;

  // let's repeat iteration until there are not too much of them ( avoid loop)
  // or the no more gain of precision is achieved
  int goon = 1;
  for( int iter=0; !iErr && goon && iter< nMaxIter; ++iter )
    {
      tmpPoint = *this;
      tmpPoint.fastswim( tarZ, iErr );

      if( iErr != 0 ){ continue; }

      // invert covariant matrix of target
      invCov = tmpPoint.cov().inverse(iErr);
      if( iErr != 0 ){ error_message(); continue; }

      // choose which parameters is used actually
      if ( coordinates ) // target is defined by coordinates
	{
	  deltaX = tmpPoint.x() - tarX;
	  deltaY = tmpPoint.y() - tarY;
	  invCov(1,1) +=icovTarXX;
	  invCov(2,2) +=icovTarYY;
	  invCov(1,2) +=icovTarXY;

	  invCov.invert(iErr);
	  if( iErr != 0 ){ error_message(); continue; }
	  
	  parNew = -invCov(iPar,1)*( deltaX * icovTarXX + deltaY*icovTarXY )
	    -       invCov(iPar,2)*( deltaY * icovTarYY + deltaX*icovTarXY );

	}else{           // target is acually a slope point 
	  deltaX = tmpPoint.tx() - tarX;
	  deltaY = tmpPoint.ty() - tarY;
	  invCov(3,3) +=icovTarXX;
	  invCov(4,4) +=icovTarYY;
	  invCov(3,4) +=icovTarXY;

	  invCov.invert(iErr);
	  if( iErr != 0 ){ error_message(); continue; }
	  
	  parNew = -invCov(iPar,3)*( deltaX * icovTarXX + deltaY*icovTarXY )
	    -       invCov(iPar,4)*( deltaY * icovTarYY + deltaX*icovTarXY );
	}

      if( fabs(parNew) < MARPLE_NEGLIGIBLE)
	{	  
	  goon = 0;
	  continue;
	}

      // the correction is very small
      if( parNew*parNew < invCov(iPar,iPar)  )
	{      
	  HepDouble difference;
	  if( fabs(tmpPoint.var(iPar) ) > MARPLE_NEGLIGIBLE )
	    difference = fabs( parNew / tmpPoint.var(iPar) );
	  else
	    difference =  fabs( parNew);	  
	  // but change of parameter is more than 10 %
	  if(  difference > 0.1 ) goon = 1; 
	  else                   
	    {
	      HepDouble xCorr = 
		-invCov(1,1)*( deltaX * icovTarXX + deltaY*icovTarXY )
		-invCov(1,2)*( deltaY * icovTarYY + deltaX*icovTarXY );
	      if( xCorr*xCorr < covTarXX )    goon = 0;
	      else                            goon = 1;
	    }
	}

      parNew += tmpPoint.var(iPar);
      setvar(iPar, parNew);
    }

  // makeTmp Point

  tmpPoint.setZ(tarZ);
  tmpPoint.setvar(iPar, parNew);
  for( unsigned int iPar2 = 0; iPar2 < 5; ++iPar2 )
    {
      if( (int)iPar2 == iPar ) continue;     
      if( coordinates )
	{
	  parNew = 
	    -invCov(iPar2,1)*( deltaX * icovTarXX + deltaY*icovTarXY )
	    -invCov(iPar2,2)*( deltaY * icovTarYY + deltaX*icovTarXY );
	}else{
	  parNew = 
	    -invCov(iPar2,3)*( deltaX * icovTarXX + deltaY*icovTarXY )
	    -invCov(iPar2,4)*( deltaY * icovTarYY + deltaX*icovTarXY );
	}
      parNew += tmpPoint.var(iPar2);
      tmpPoint.setvar(iPar2, parNew);
    }
  if( coordinates)
    {
      tmpPoint.setcov(1,1, covTarXX );
      tmpPoint.setcov(1,2, covTarXY );
      tmpPoint.setcov(2,2, covTarYY );
      for( unsigned int ic = 1; ic<=5; ++ic)
	for( unsigned int jc = ic; jc<=5; ++jc)
	  {
	    if( ic>2 || jc>2 )
	      {
		if( ic == jc)  tmpPoint.setcov(ic,jc,MARPLE_INFINITY); 
		else           tmpPoint.setcov(ic,jc,0);
	      }
	  }
    }else{
      tmpPoint.setcov(3,3, covTarXX );
      tmpPoint.setcov(3,4, covTarXY );
      tmpPoint.setcov(4,4, covTarYY );
      for( unsigned int ic = 1; ic<=5; ++ic)
	for( unsigned int jc = ic; jc<=5; ++jc)
	  {
	    if( ic<3 || ic > 4 || jc<3 || jc > 4  )
	      {
		if( ic == jc)  tmpPoint.setcov(ic,jc,MARPLE_INFINITY); 
		else           tmpPoint.setcov(ic,jc,0);
	      }
	  }
    }
  
  // now we put estimation as initial guess, prolongate it 
  // to original point and make fit
  // which is just the same match on everything
 
  if( iErr )
    {
# ifdef MARPLEPERFORMANCES
      chRcPointEstimate->stop();
# endif
      return ;
    }
  tmpPoint.prolongate(zOld, iErr);
  if( iErr )
    {
# ifdef MARPLEPERFORMANCES
  chRcPointEstimate->stop();
# endif
      return ;
    }
  RcMatchPoint makeMatch( *this, tmpPoint, 11111);

  operator = (makeMatch);

# ifdef MARPLEPERFORMANCES
  chRcPointEstimate->stop();
# endif

}


//----------------------------------------------------------------

extern void maFindTarget( RcPoint & rp, 
			  HepDouble &tarX,   HepDouble &tarY,  
			  HepDouble &tarZ,
			  HepDouble &covTarXX, HepDouble &covTarYY, 
			  HepDouble &covTarXY, 
			  int & iErr);

//----------------------------------------------------------------

void RcPoint::estimatePEfromTAR( int nMaxIter, 
				 int & iErr )
{
  estimatePARfromTAR( 5,  nMaxIter, iErr );
}

void RcPoint::estimatePARfromTAR(  int iPar,
				   int nMaxIter, 
				   int & iErr )
{
# ifdef MARPLEPERFORMANCES
  if(chRcPointEstimate == 0){
    ChronoManager CM;
    chRcPointEstimate = CM.createChrono("estimatePar ");
  }
  chRcPointEstimate->start();
# endif

  HepDouble zOld = z();

  HepDouble parNew = 0;
  HepDouble tarX, tarY;
  HepDouble covTarXX, covTarYY, covTarXY;
  HepDouble tarZ =0;
  HepDouble deltaX = 0;
  HepDouble deltaY = 0;
  HepDouble icovTarXX = 0;
  HepDouble icovTarYY = 0;
  HepDouble icovTarXY = 0;
  iErr = 0;
  int goon = 1;
  HepSymMatrix invCov(5,1);
  RcPoint tmpPoint;

  for( int iter=0; !iErr && goon && iter< nMaxIter; iter++ )
    {
      tmpPoint = *this;
      //tmpPoint.fastswim( tarZ, iErr );
      tmpPoint.prolongate( tarZ, iErr );

      if( iErr != 0 ){ continue; }

      // to be replaced by VERTEX subroutine/function !!!
      maFindTarget( tmpPoint, tarX, tarY, tarZ, 
		    covTarXX, covTarYY, covTarXY, 
		    iErr );
      if( iErr ) continue;
      tmpPoint.lineswim( tarZ);

      // invert covariant matrix of target
      if( covTarXY == 0 )
	{
	  if( covTarXX ) icovTarXX = 1 / covTarXX; else icovTarXX = 0;
	  if( covTarYY ) icovTarYY = 1 / covTarYY; else icovTarYY = 0;
	  icovTarXY = 0;
	}else{
	  HepDouble determinant = 1/(covTarXX*covTarYY-covTarXY*covTarXY);
	  if( fabs(determinant) < MARPLE_NEGLIGIBLE ) determinant = 1;
	  icovTarXY = -covTarXY / determinant;
	  HepDouble tmpXX = covTarXX;
	  icovTarXX = covTarYY/determinant; // invertion
	  icovTarYY = tmpXX/determinant;
	}
      

      invCov = tmpPoint.cov().inverse(iErr);
      if( iErr ){ error_message(); continue; }

      invCov(1,1) +=icovTarXX;  // icovTar** are already inverted
      invCov(2,2) +=icovTarYY;
      invCov(1,2) +=icovTarXY;
      
      invCov.invert(iErr);

      if( iErr ){ error_message();  continue; }

      deltaX = tmpPoint.x() - tarX;
      deltaY = tmpPoint.y() - tarY;

      parNew = -invCov(iPar,1)*( deltaX * icovTarXX + deltaY*icovTarXY )
	-       invCov(iPar,2)*( deltaY * icovTarYY + deltaX*icovTarXY );
      
      if( fabs(parNew) < MARPLE_NEGLIGIBLE)
	{	  
	  goon = 0;
	  parNew += tmpPoint.var(iPar);
	  setvar(iPar, parNew);     
	  continue;
	}
      // the correction is very small
      if( parNew*parNew < invCov(iPar,iPar)/2 )
	{      
	  HepDouble difference;
	  if( fabs(tmpPoint.var(iPar) ) > MARPLE_NEGLIGIBLE )
	    difference = fabs( parNew / tmpPoint.var(iPar) );
	  else
	    difference =  0;	  
	  // but change of parameter is more than 5 %
	  if(  difference > 0.05 ) goon = 1; 
	  else
	    {
	      HepDouble xCorr = 
		-invCov(1,1)*( deltaX * icovTarXX + deltaY*icovTarXY )
		-invCov(1,2)*( deltaY * icovTarYY + deltaX*icovTarXY );
	      if( xCorr*xCorr < covTarXX/2 )    goon = 0;
	      else                              goon = 1;
	    }
	}
      parNew += tmpPoint.var(iPar);
      setvar(iPar, parNew);     
    }

  if( iErr )
    {
# ifdef MARPLEPERFORMANCES
      chRcPointEstimate->stop();
# endif
      return ;
    }
  
  tmpPoint.setZ(tarZ);
  tmpPoint.setvar(iPar, parNew);
  tmpPoint.setvar(1,tarX);
  tmpPoint.setvar(2,tarY);
  for( unsigned int iPar2 = 3; iPar2 <= 5; ++iPar2 )
    {
      if( (int)iPar2 == iPar ) continue;     
      parNew = 
	-invCov(iPar2,1)*( deltaX * icovTarXX + deltaY*icovTarXY )
	-invCov(iPar2,2)*( deltaY * icovTarYY + deltaX*icovTarXY );
      parNew += tmpPoint.var(iPar2);
      tmpPoint.setvar(iPar2, parNew);
    }
  for( unsigned int ic = 1; ic<=5; ++ic)
    for( unsigned int jc = ic; jc<=5; ++jc)
      {
	if( ic>2 || jc>2 )
	  {
	    if( ic == jc) tmpPoint.setcov(ic,jc,1.); 
	    else           tmpPoint.setcov(ic,jc,0);
	  }
      }

  tmpPoint.setcov(1,1, covTarXX);
  tmpPoint.setcov(1,2, covTarXY);
  tmpPoint.setcov(2,2, covTarYY);

  tmpPoint.setcov(5,5,50);

  /*   float resol = tmpPoint.pe();
     if( fabs(resol)<0.002 ) resol = 0.002;
     resol *= resol*100; 
     tmpPoint.setcov(5,5,resol);
     */  

  // now we put estimation as initial guess, prolongate it 
  // to original point and make fit
  // which is just the usual match on everything
 
  if( iErr )
    {
# ifdef MARPLEPERFORMANCES
      chRcPointEstimate->stop();
# endif
      return ;
    }
  tmpPoint.prolongate(zOld, iErr);
  if( iErr )
    {
# ifdef MARPLEPERFORMANCES
      chRcPointEstimate->stop();
# endif
      return ;
    }

  RcMatchPoint makeMatch( *this, tmpPoint, 11111);
  operator = (makeMatch);
  
# ifdef MARPLEPERFORMANCES
  chRcPointEstimate->stop();
# endif

}


//-----------------------------------------------------------------------
// returns empty mtraList
vector<ArtePointer<MTRA> > RcPoint::mtraList() const
{
  vector<ArtePointer<MTRA> > dummy;
  return dummy;
}

// returns empty hitbList
vector<ArtePointer<HITB> > RcPoint::hitbList() const
{
  vector<ArtePointer<HITB> > dummy;
  return dummy;
}
// returns empty hitcList
vector<ArtePointer<HITC> > RcPoint::hitcList() const
{
  vector<ArtePointer<HITC> > dummy;
  return dummy;
}

//-----------------------------------------------------------------------
// update ARTE bank connected with point
void RcPoint::updateARTE(int & iErr ) const
{
  iErr = 1;
  arteMsg << "MARPLE: updateARTE : ARTE bank is unknown";
  arteMsg.allp();
}


//-----------------------------------------------------------------------
// return 1 if at least one element in mtraList1 == any element in mtraList2
int RcMtraOk( const vector<ArtePointer<MTRA> > & mtraList1, 
	      const vector<ArtePointer<MTRA> > & mtraList2 )
{
  if( RcMtraPointer(mtraList1, mtraList2) != 0 ) return 1;
  return 0;
}


ArtePointer<MTRA> 
RcMtraPointer( const vector<ArtePointer<MTRA> > & mtraList1, 
	       const vector<ArtePointer<MTRA> > &  mtraList2 )
{
  for( unsigned int iMtra1 = 0; 
       iMtra1 < mtraList1.size(); ++iMtra1 )
    {
      for( unsigned int iMtra2 = 0; 
	   iMtra2 < mtraList2.size(); ++iMtra2 )
	{
	  if( mtraList1[iMtra1] == mtraList2[iMtra2] )
	    {  return  mtraList1[iMtra1] ; }
	}
    }
  return 0;
}

int RcMtraOk( ArtePointer<MTRA>  & mtraEntry, 
	      const vector<ArtePointer<MTRA> > & mtraList )
{
  if( RcMtraPointer(mtraEntry, mtraList ) != 0 ) return 1;
  return 0;
}

ArtePointer<MTRA> RcMtraPointer( ArtePointer<MTRA>  & mtraEntry, 
				 const vector<ArtePointer<MTRA> > & mtraList )
{
  for( unsigned int iMtra = 0; 
       iMtra < mtraList.size(); iMtra++ )
    {
      if( mtraEntry == mtraList[iMtra] )
	    {   return  mtraList[iMtra]; }
    }
  return 0;
}


//-----------------------------------------------------------------------
//return 1 if RcPoint leaved by lepton from J/psi
int RcFromJpsi( const RcPoint & rp)
{
  vector<ArtePointer<MTRA> > mtraList = rp.mtraList();
  for( unsigned int iMtra = 0; 
       iMtra < mtraList.size(); iMtra++ )
    {
      ArtePointer<MTRA> mtra = mtraList[iMtra];
      int iLund = mtra->topp->lund;
      if( abs(iLund) == 11 || abs(iLund) == 13 )
	{
	  mtra  = mtra->btra;
	  iLund = mtra->topp->lund;
	  if( abs(iLund) == 443 )
	    {  return 1;  }
	}
    }
  return 0;
}

//-----------------------------------------------------------------------
// print ostream

ostream & operator << (ostream & os, const RcPoint &rp)
{
  os << " z= " << setw(4) << rp.z();
  os << "; x = " <<  setw(4) << rp.x() ;
  os << ";  y = " << setw(4) << rp.y() ;
  os << "; tx = " << setw(4) << rp.tx() ;
  os << "; ty = " << setw(4) << rp.ty() ;
  os << "; pe = " << setw(4) << rp.pe() <<endl ;
  os << rp.cov();
  os << endl;
  os << "         marking flag = " << rp.mark() << endl;
  os <<  "            MTRA link: " ;
  vector<ArtePointer<MTRA> > mList( rp.mtraList());
  unsigned int i=0;
  for(i=0; i < mList.size(); i++)
    {
      os << mList[i];
    }
  os << endl;
  os <<  "            HITB link: " ;
  vector<ArtePointer<HITB> > hbList( rp.hitbList());
  for( i=0; i < hbList.size(); i++)
    {
      os << hbList[i];
    }
  os << endl;
  os <<  "            HITC link: " ;
  vector<ArtePointer<HITC> > hcList( rp.hitcList());
  for(i=0; i < hcList.size(); i++)
    {
      os << hcList[i];
    }
  os << endl;

  return os;
}

// ostream for mtraList

ostream & operator << (ostream & os, 
		       const vector<ArtePointer<MTRA> > mtraList)
{
  for(unsigned int i=0; i < mtraList.size(); i++ )
    {  os <<mtraList[i]; }
  return os;
}
