//--------------------------------------------------------------------
// Created: 02.03.98 by   Olya Igonkina

/* MARPLE function for matching of SVD tracks and ECAL clusters 
   for case when Magnet is switched OFF. 
   
   Usage :
      extern "C"  
      void FORTRAN_ROUTINE(matchsenomag)( float * threshold_cut ) 

   Parameters:
      threshsold_cut is a cut on Energy of cluster,
      (if threshold_cut > 200 - it is Ktrig cut, otherwise it is pt_cut )

   Called: 
      by matchsvdecal_() or can be called directly
   */

//--------------------------------------------------------------------
#include <math.h>

#include "arte/RSEG_cmp.hh"

#include "marple/marple.hh"
#include "marple/RcSegClusBuilder.hh"

//definition of constants for SUSI-CARE matching ( see \Ref{RcMatchConst})
// flag, z_plane, z_min, z_max, use_pair_cut, n_pair_left, n_pair_right, 
//use_delta_cut, delta_x, delta_y, delta_tx, delta_ty, delta_pe, chi2

// match with Ecluster

RcMatchConst rcSEnoMagConst = { 
 11000,     1330.,   -30.,  1500.,      1,          1,            1,
     2,       3.,      3.,   1.e+10,  1.e+10,   1.e+10,   60.
};

RcMatchConst rcSESoftnoMagConst = { 
 11000,     1330.,   -30.,  1500.,      1,          1,            1,
     2,       15.,      15.,   1.e+10,  1.e+10,   1.e+10,   100.
};

// if threshold_cut > 200 - it is Ktrig cut, otherwise it is pt_cut  

extern "C"  
void FORTRAN_ROUTINE(matchsenomag)( float * threshold_cut ) 
{
  RcSegClusBuilder SvdEcalNoMag;

  //--------------------------------------------------------------------
  // put constants to the MatchBuilder
  //--------------------------------------------------------------------
  if( MARPLE_SOFT )
    {
      SvdEcalNoMag.setConst(rcSESoftnoMagConst);
      SvdEcalNoMag.setBremConst(rcSESoftnoMagConst);
    }else{
      SvdEcalNoMag.setConst(rcSEnoMagConst);
      SvdEcalNoMag.setBremConst(rcSEnoMagConst);
    }
  //--------------------------------------------------------------------
  // create vector of clusters from CARE
  //--------------------------------------------------------------------

  for( ArteTable<RCCL>::iterator irccl = ArteTable<RCCL>::begin();
       irccl != ArteTable<RCCL>::end(); ++irccl )
    {
      if( (*irccl)->flag < 0 ) continue;
      if( maTakeRccl( *threshold_cut, *irccl) )
	{
	  RcClusPoint * ptPoint = new RcClusPoint( (*irccl));
	  SvdEcalNoMag.addPointRight( *ptPoint);
	  ptPoint = 0;
	}
    }  

  //--------------------------------------------------------------------
  // create vector of segments from SVD
  //--------------------------------------------------------------------
  
  for( ArteTable<RSEG>::iterator irseg = ArteTable<RSEG>::begin();
       irseg != ArteTable<RSEG>::end(); ++irseg )
    {
      if( //(*irseg)->fit > 0 && 
	  (*irseg)->cmp == Rsegc::vxd && 
	  SvdEcalNoMag.ptToMatchConst()->geomAccept(double((*irseg)->ze)) )
	{ 	  
	  RcSegPoint * ptToPoint = new RcSegPoint( (*irseg) , 1); // endPoint
	  ptToPoint->lineswim(SvdEcalNoMag.ptToMatchConst()->z_plane);
	  SvdEcalNoMag.addPointLeft(  *ptToPoint);
	  ptToPoint = 0;
	}
    }
  //--------------------------------------------------------------------
  // do Matching 
  //--------------------------------------------------------------------

  int iErr = 0;

  SvdEcalNoMag.doMatch();

  //--------------------------------------------------------------------
  // outprint
  //--------------------------------------------------------------------

  if( MARPLE_DEBUG >= 0)
    {
      SvdEcalNoMag.statMC();  // check MC information  
      
      if( MARPLE_DEBUG >= 20)
	{
	  SvdEcalNoMag.fillMatch(8116);
	  SvdEcalNoMag.fillDeltas(8216, 8216);
	}else  if( MARPLE_DEBUG >= 10){
	  SvdEcalNoMag.fillDeltas(8216);
	}
      static int firstSVDCARE = 1;
      if( firstSVDCARE )
	{
	  cout << *SvdEcalNoMag.ptToMatchConst();
	  firstSVDCARE = 0;
	}
      
      // print result of matching
      SvdEcalNoMag.stat( MARPLE_DEBUG, 1);
    }
  
  //--------------------------------------------------------------------
  // save found relation to the table
  //--------------------------------------------------------------------
  if( MARPLE_MAKERSEG ) SvdEcalNoMag.saveRSEG(iErr);

  SvdEcalNoMag.saveRelation(iErr);
  if( iErr ){ cout << "MARPLE: some relations were not saved 1 " << endl;}

  //--------------------------------------------------------------------
  // clean everything
  //--------------------------------------------------------------------
  SvdEcalNoMag.deleteAll();  
}

