/* Usage:

   call_via_istr_1par              (F f, (Arg1*)0, ..., (ArgN*)0, (P1*)0); or
   call_via_istr_check_EOF_1par    (F f, (Arg1*)0, ..., (ArgN*)0, (P1*)0); or
   call_via_istr_conditionally_1par(F f, (Arg1*)0, ..., (ArgN*)0, (P1*)0,
                                    Condition c);

   where F f - some function object which takes N+1 (0<=N<=5) arguments and
   expression f(a1, .., aN, p1)
   (a1, .., aN are of types Arg1, .., ArgN respectively, p1 is of type P1)
   is valid.
   If call_*_conditionally_1par(..) is used: Condition c is such an object
   that the expression c(is) is valid ("is" is of type istream*).

   call_via_istr*(...) returns an object of type CallViaIstr0..5_1Par<...>
   which has the operator()(istream* is, P1 p1) => it extracts N arguments
   from "*is",
   if they are extracted successfully ("*is" state is good), "Condition c"
   is applied to "is" and:
   if ( c(is) ) => function f(a1, .., aN, p1) is called.
   In case of simple call_via_istr_1par(..) version condition is substituted
   by the object which always returns true, so that condition has no effect;
	 call_via_istr_check_EOF_1par(..) has a condition, which checks that after
	 extracting of all arguments (*is) is empty.
*/

// Author V.Balagura (balagura@mail.desy.de)   (16.04.98).

#ifndef _call_via_istr_1par_hh
#define _call_via_istr_1par_hh 1

#include "ptset/call_via_istr.hh"

/* -------------------- 0 args -------------------- */
template <class F, class P1, class Condition>
class CallViaIstr0_1Par {
public:
  F function;
  Condition condition;

  CallViaIstr0_1Par(F f, Condition c = Condition())
	 : function(f), condition(c) {}
  void operator()(istream* is, P1 p1) {
    if ( condition(is) ) function(p1);
  }
};
template <class F, class P1, class Condition>
inline CallViaIstr0_1Par<F, P1, Condition> call_via_istr_conditionally_1par
(F f, P1*, Condition c) {
  return CallViaIstr0_1Par<F, P1, Condition>(f, c);
}
/* -------------------- 1 arg -------------------- */
template <class F, class A1, class P1, class Condition>
class CallViaIstr1_1Par {
public:
  F function;
  Condition condition;

  CallViaIstr1_1Par(F f, Condition c = Condition())
	 : function(f), condition(c) {}
  void operator()(istream* is, P1 p1) {
    A1 a1;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract argument"); return; }
    if ( condition(is) ) function(a1, p1);
    else global_error() << a1;
  }
};
template <class F, class A1, class P1, class Condition>
inline CallViaIstr1_1Par<F, A1, P1, Condition>
call_via_istr_conditionally_1par
(F f, A1*, P1*, Condition c) {
  return CallViaIstr1_1Par<F, A1, P1, Condition>(f, c);
}
/* -------------------- 2 args -------------------- */
template <class F, class A1, class A2, class P1, class Condition>
class CallViaIstr2_1Par {
public:
  F function;
  Condition condition;

  CallViaIstr2_1Par(F f, Condition c = Condition())
	 : function(f), condition(c) {}
  void operator()(istream* is, P1 p1) {
    A1 a1;
    A2 a2;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract first argument"); return; }
    *is >> a2;
    if ( !(*is) )
      { global_error()("can not extract second argument"); global_error() << a1; return; }
    if ( condition(is) ) function(a1, a2, p1);
    else global_error() << a2 << a1;
  }
};
template <class F, class A1, class A2, class P1, class Condition>
inline CallViaIstr2_1Par<F, A1, A2, P1, Condition>
call_via_istr_conditionally_1par
(F f, A1*, A2*, P1*, Condition c) {
  return CallViaIstr2_1Par<F, A1, A2, P1, Condition>(f, c);
}
/* -------------------- 3 args -------------------- */
template <class F, class A1, class A2, class A3, class P1, class Condition>
class CallViaIstr3_1Par {
public:
  F function;
  Condition condition;

  CallViaIstr3_1Par(F f, Condition c = Condition())
	 : function(f), condition(c) {}
  void operator()(istream* is, P1 p1) {
    A1 a1;
    A2 a2;
    A3 a3;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract first argument"); return; }
    *is >> a2;
    if ( !(*is) )
      { global_error()("can not extract second argument"); global_error() << a1; return; }
    *is >> a3;
    if ( !(*is) )
      { global_error()("can not extract third argument"); global_error() << a2 << a1; return; }
    if ( condition(is) ) function(a1, a2, a3, p1);
    else global_error() << a3 << a2 << a1;
  }
};
template <class F, class A1, class A2, class A3, class P1, class Condition>
inline CallViaIstr3_1Par<F, A1, A2, A3, P1, Condition>
call_via_istr_conditionally_1par
(F f, A1*, A2*, A3*, P1*, Condition c) {
  return CallViaIstr3_1Par<F, A1, A2, A3, P1, Condition>(f, c);
}
/* -------------------- 4 args -------------------- */
template <class F, class A1, class A2, class A3, class A4, class P1,
  class Condition>
class CallViaIstr4_1Par {
public:
  F function;
  Condition condition;

  CallViaIstr4_1Par(F f, Condition c = Condition())
	 : function(f), condition(c) {}
  void operator()(istream* is, P1 p1) {
    A1 a1;
    A2 a2;
    A3 a3;
    A4 a4;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract first argument"); return; }
    *is >> a2;
    if ( !(*is) )
      { global_error()("can not extract second argument"); global_error() << a1; return; }
    *is >> a3;
    if ( !(*is) )
      { global_error()("can not extract third argument"); global_error() << a2 << a1; return; }
    *is >> a4;
    if ( !(*is) ) {
      global_error()("can not extract fourth argument"); global_error() << a3 << a2 << a1;
      return;
    }
    if ( condition(is) ) function(a1, a2, a3, a4, p1);
    else global_error() << a4 << a3 << a2 << a1;
  }
};
template <class F, class A1, class A2, class A3, class A4, class P1,
  class Condition>
inline CallViaIstr4_1Par<F, A1, A2, A3, A4, P1, Condition>
call_via_istr_conditionally_1par
(F f, A1*, A2*, A3*, A4*, P1*, Condition c) {
  return CallViaIstr4_1Par<F, A1, A2, A3, A4, P1, Condition>(f, c);
}

/* -------------------- 5 args -------------------- */
template <class F, class A1, class A2, class A3, class A4, class A5, class P1,
  class Condition>
class CallViaIstr5_1Par {
public:
  F function;
  Condition condition;

  CallViaIstr5_1Par(F f, Condition c = Condition())
	 : function(f), condition(c) {}
  void operator()(istream* is, P1 p1) {
    A1 a1;
    A2 a2;
    A3 a3;
    A4 a4;
    A5 a5;
    *is >> a1;
    if ( !(*is) ) { global_error()("can not extract first argument"); return; }
    *is >> a2;
    if ( !(*is) )
      { global_error()("can not extract second argument"); global_error() << a1; return; }
    *is >> a3;
    if ( !(*is) )
      { global_error()("can not extract third argument"); global_error() << a2 << a1; return; }
    *is >> a4;
    if ( !(*is) ) {
      global_error()("can not extract fourth argument"); global_error() << a3 << a2 << a1;
      return;
    }
    *is >> a5;
    if ( !(*is) ) {
      global_error()("can not extract fifth argument"); global_error() << a4 << a3 << a2 << a1;
      return;
    }
    if ( condition(is) ) function(a1, a2, a3, a4, a5, p1);
    else global_error() << a5 << a4 << a3 << a2 << a1;
  }
};
template <class F, class A1, class A2, class A3, class A4, class A5, class P1,
  class Condition>
inline CallViaIstr5_1Par<F, A1, A2, A3, A4, A5, P1, Condition>
call_via_istr_conditionally_1par
(F f, A1*, A2*, A3*, A4*, A5*, P1*, Condition c) {
  return CallViaIstr5_1Par<F, A1, A2, A3, A4, A5, P1, Condition>(f, c);
}

/* -------------------- call_via_istr_1par -------------------- */
/* No condition on istream is required */

template <class F, class P1>
inline CallViaIstr0_1Par<F, P1, _IstrOk> call_via_istr_1par(F f, P1*) {
  return CallViaIstr0_1Par<F, P1, _IstrOk>(f, _IstrOk());
}
template <class F, class A1, class P1>
inline CallViaIstr1_1Par<F, A1, P1, _IstrOk>
call_via_istr_1par(F f, A1*, P1*) {
  return CallViaIstr1_1Par<F, A1, P1, _IstrOk>(f, _IstrOk());
}
template <class F, class A1, class A2, class P1>
inline CallViaIstr2_1Par<F, A1, A2, P1, _IstrOk>
call_via_istr_1par(F f, A1*, A2*, P1*) {
  return CallViaIstr2_1Par<F, A1, A2, P1, _IstrOk>(f, _IstrOk());
}
template <class F, class A1, class A2, class A3, class P1>
inline CallViaIstr3_1Par<F, A1, A2, A3, P1, _IstrOk>
call_via_istr_1par(F f, A1*, A2*, A3*, P1*) {
  return CallViaIstr3_1Par<F, A1, A2, A3, P1, _IstrOk>(f, _IstrOk());
}
template <class F, class A1, class A2, class A3, class A4, class P1>
inline CallViaIstr4_1Par<F, A1, A2, A3, A4, P1, _IstrOk>
call_via_istr_1par(F f, A1*, A2*, A3*, A4*, P1*) {
  return CallViaIstr4_1Par<F, A1, A2, A3, A4, P1, _IstrOk>(f, _IstrOk());
}
template <class F, class A1, class A2, class A3, class A4, class A5, class P1>
inline CallViaIstr5_1Par<F, A1, A2, A3, A4, A5, P1, _IstrOk>
call_via_istr_1par(F f, A1*, A2*, A3*, A4*, A5*, P1*) {
  return CallViaIstr5_1Par<F, A1, A2, A3, A4, A5, P1, _IstrOk>(f, _IstrOk());
}

/* ----------------- call_via_istr_check_EOF_1par ----------------- */
/* Condition: istream is empty */

template <class F, class P1>
inline CallViaIstr0_1Par<F, P1, _CheckIstr_EOF>
call_via_istr_check_EOF_1par(F f, P1*) {
  return CallViaIstr0_1Par<F, P1, _CheckIstr_EOF>(f, _CheckIstr_EOF());
}
template <class F, class A1, class P1>
inline CallViaIstr1_1Par<F, A1, P1, _CheckIstr_EOF>
call_via_istr_check_EOF_1par(F f, A1*, P1*) {
  return CallViaIstr1_1Par<F, A1, P1, _CheckIstr_EOF>(f, _CheckIstr_EOF());
}
template <class F, class A1, class A2, class P1>
inline CallViaIstr2_1Par<F, A1, A2, P1, _CheckIstr_EOF>
call_via_istr_check_EOF_1par(F f, A1*, A2*, P1*) {
  return CallViaIstr2_1Par<F, A1, A2, P1, _CheckIstr_EOF>
	  (f, _CheckIstr_EOF());
}
template <class F, class A1, class A2, class A3, class P1>
inline CallViaIstr3_1Par<F, A1, A2, A3, P1, _CheckIstr_EOF>
call_via_istr_check_EOF_1par(F f, A1*, A2*, A3*, P1*) {
  return CallViaIstr3_1Par<F, A1, A2, A3, P1, _CheckIstr_EOF>
	  (f, _CheckIstr_EOF());
}
template <class F, class A1, class A2, class A3, class A4, class P1>
inline CallViaIstr4_1Par<F, A1, A2, A3, A4, P1, _CheckIstr_EOF>
call_via_istr_check_EOF_1par(F f, A1*, A2*, A3*, A4*, P1*) {
  return CallViaIstr4_1Par<F, A1, A2, A3, A4, P1, _CheckIstr_EOF>
	  (f, _CheckIstr_EOF());
}
template <class F, class A1, class A2, class A3, class A4, class A5, class P1>
inline CallViaIstr5_1Par<F, A1, A2, A3, A4, A5, P1, _CheckIstr_EOF>
call_via_istr_check_EOF_1par(F f, A1*, A2*, A3*, A4*, A5*, P1*) {
  return CallViaIstr5_1Par<F, A1, A2, A3, A4, A5, P1, _CheckIstr_EOF>
	  (f, _CheckIstr_EOF());
}

#endif
