/* 
   Defines class GlobalError.
   This class was developed since compilers did not support exception handling.
-------------------------------------------------------------------------------
     GlobalError& global_error(); 
   function returns object of type GlobalError which is
   used everywhere in my codes. If global_error()() == "" (no error message)
   there were no errors. If error occurs a message is written to
   global_error() like
     string tmp;
     tmp << "index out of range: " << i << " > " << i_max;
     global_error() << tmp;
   It is possible to put anything in the message which can be written to cout,
   i.e. 
     global_error() << 3.14;
   Suppose the error occured in function f() which was called from function
   g(). f() has written something to global_error() signalling about the
   error. 
   In g() one may check 
     if (global_error()() != "") {
   and if there was an error, add to the error message some information which
   is available at g() level:
       global_error() << "I (g()) called f() and this caused an error";
       return;
     }
   If, e.g. in the main program you'd like to call g() and if there is an error
   print the error message to cerr:
     int main() {
       g();
       if ( global_error() != "") {
         cerr << global_error();
	 return 1;
       }
       ...
     }
     This will print FIRST the message from g(), and THEN message from f().
     Note that the order is opposite to the order of placing the messages
     to global_error(): first message has been written by f(), second - by g().

     Every next global_error() << ...; will produce a line in the output
     which will be BEFORE all the next messages. 
     Therefore do not use e.g. 
       global_error() << "index out of range: " << i << " > " << i_max;
     since every part of the message will be on a separate line in
     reverse order. 
     */

// Author V.Balagura (balagura@mail.desy.de)   (23.03.98).

#ifndef _global_error_hh
#define _global_error_hh 1

#include <string>
#include <strstream.h>
#include "ptset/string_primitives.hh"

class GlobalError {
  string message;
  int indent;
  friend ostream& operator<<(ostream& os, const GlobalError& er) {
    return os << "*** error:\n" << er.message;
  }
public:
  GlobalError() : indent(0) {}
  void operator()(const string& s) {
    string indent_(indent, ' ');
    message = indent_ + s + '\n';
  }
  string operator()() {
    if ( message.size() != 0 )
      return string(message, 0, message.size()-1);  // remove last '\n'
    else return "";
  }
  void clear() { message = ""; }
  bool operator!() { return message == ""; }
  operator void*() { return (message == "" ? NULL : (char*)(NULL) + 1); }
  // no indent
  void prepend(const string& s) { message += s; message += '\n'; }
  void prepend_with_indent(const string& s) {
    string indent_(indent, ' ');
    message += indent_;
    message += s;
    message += '\n';
  }
  void increment_indent() { indent += 2; }
  void decrement_indent() { indent -= 2; }
};
// with indent
template<class T>
inline GlobalError& operator<<(GlobalError& e, const T& x) {
  string tmp;
  tmp << x;
  e.prepend_with_indent(tmp);
  return e;
}
inline GlobalError& operator<<(GlobalError& e, const char* x) {
  e.prepend_with_indent(string(x));
  return e;
}
inline GlobalError& operator<<(GlobalError& e, const string& x) {
  e.prepend_with_indent(x);
  return e;
}
  

GlobalError& global_error();

#endif
