#include "rich/vessel.h"
#include "rich/Losses.h"

extern int RICHdebug;

// ===========================================================================
//
//  reflect: search single mirror and execute reflexion
//

int PlaneMirror::reflect(Photon* pp) {
  if (speed==1) return fullmirror_reflect(pp);
  int ret;
  double path;
  Photon& p=*pp;

  if ((ret=fullmirror_intersection(p,&path))) return ret;

  Vec3 x=p.x()+path*p.n();

  int k;
  if (x.y()>0) {
    x=fullupper.localhit();
    k=index(3+fullupper.row(x.y()),fullupper.col(x.x()));
  }
  else {
    x=fulllower.localhit();
    k=index(fulllower.row(x.y()),fulllower.col(x.x()));
  }
  if ((ret=mirrors[k]->reflect(&p))) return ret;

  x=RICH_Vessel->ElectronBeamShroud().TransformFromOld(p.x());
  if (x.x()*x.x()+x.y()*x.y()<RadiusElectronHole()*RadiusElectronHole()) {
    return PMI_LOSS_EBEAMCUTOUT;
    }
  return NO_LOSS;  
  }

// Get plane mirror
SinglePlaneMirror *PlaneMirror::GetSingleMirror(Vec3 x) {
  int k;
  if (x.y()>0) {
    x=x-fullupper.x0();
    k=index(3+fullupper.row(x.y()),fullupper.col(x.x()));
  }
  else {
    x=fulllower.localhit();
    k=index(fulllower.row(x.y()),fulllower.col(x.x()));
  }
  return mirrors[k];
}

// -----------------------------------------------------------------------
//
//  find intersection with full mirror
//
int PlaneMirror::fullmirror_intersection(const StraightLine& p,double *path){

  if (p.x().y()>0&&p.n().y()>=0) {
    if (fullupper.intersection(p,path))    return PMI_LOSS_OUTOFMIRROR;
  }
  else if (p.x().y()<0&&p.n().y()<=0) {
    if (fulllower.intersection(p,path)) return PMI_LOSS_OUTOFMIRROR;
  }
  else if (p.x().y()>0) {
    if (fullupper.intersection(p,path)) {
      if (fulllower.intersection(p,path)) return PMI_LOSS_OUTOFMIRROR;
    }
  }   
  else  {
    if (fulllower.intersection(p,path)) {
      if (fullupper.intersection(p,path)) return PMI_LOSS_OUTOFMIRROR;
    }
  }
  return NO_LOSS;
}

// -----------------------------------------------------------------------
//
//  used to estimate the geometrical efficiency on track by track basis
//  and as part of the fast simulation
//

int PlaneMirror::fullmirror_reflectonly(Photon* pp) {
  int ret;
  double path;
  Photon &p=*pp;
  
  if ((ret=fullmirror_intersection(p,&path))) return ret;
  if ((ret=RICH_Vessel->traceonly(&p,path))) return ret;

  Vec3 y=RICH_Vessel->ElectronBeamShroud().TransformFromOld(p.x());
  if (y.x()*y.x()+y.y()*y.y()<RadiusElectronHole()*RadiusElectronHole()) {
    return PMI_LOSS_EBEAMCUTOUT;
    }

  Vec3 ns=fulllower.ez();
  if (p.x().y()>0) ns=fullupper.ez();
  
  p.n()-=double(2)*(ns*p.n())*ns;     // short hand formula for the reflexion    

  return NO_LOSS;
  }

// -----------------------------------------------------------------------
//
//  reflexion on full mirror used for fast simulation: no losses in mirror gaps
//


int PlaneMirror::fullmirror_reflect(Photon* pp) {
  int ret;
  if ((ret=fullmirror_reflectonly(pp))) return ret;
  if (rnd_normal()>reflectivity().f(pp->lambda())) return PMI_LOSS_REFLECTIVITY;
  return NO_LOSS;
  }  
 
// -----------------------------------------------------------------------
//
//   print data
//

void PlaneMirror::print(FILE *fp) {
  fprintf(fp,"#\nPlaneMirror\n");
  fprintf(fp,"  numberofmirrors   %d\n",nummirror);
  fprintf(fp,"  radius_electron_hole %f\n",radiuselectronhole);
  fprintf(fp,"#\n reflectivity\n");
  refl.print(fp,0);
  fprintf(fp,"  end  # plane mirror \n");
  int r,c;
  for (r=0;r<fullupper.nx();r++) 
    for (c=0;c<fullupper.ny()*2;c++) 
      SingleMirror(r,c)->print(fp);
  }

// -----------------------------------------------------------------------
//
//  initialisation from file
//

PlaneMirror::PlaneMirror(ScriptFile& f) {
  char n[256],tag[256];
  
  FullSimulation();
  while (f.getline(n)) {
    char *s=n;
    sscanf(s,"%s",tag);s+=f.itemlength(s,1);
    int err=0;
    if (!strcmp(tag,"numberofmirrors"))   
      err=(sscanf(s,"%d",&nummirror)!=1);  
    else if (!strcmp(tag,"radius_electron_hole")) {
      err=(sscanf(s,"%lf",&radiuselectronhole)!=1); 
      } 
    else if (!strcmp(tag,"reflectivity"))   
      refl=Table(f);  
    else if (!strcmp(tag,"end"))  break;
    else {
      fprintf(stderr,
         "PlaneMirror(ScriptFile):  Unknown tag: %s err=%d\n",tag,err);
      Exit(0);
      }
    }

 

  nnumcol=6;
  mirrors=new SinglePlaneMirror*[nummirror];

  int i;
  SinglePlaneMirror* m;
  for (i=0;i<nummirror+2;i++) {
    f.getline(n);sscanf(n,"%s",tag);
    if (strcmp(tag,"SinglePlaneMirror")) {
      fprintf(stderr,"PlaneMirror(ScriptFile): Not enough single planar mirrors are"
              " defined in the input file. Abort\n");
      Exit(1);
      }
    m=new SinglePlaneMirror(f,refl);
    if (RICHdebug) m->print();
    if (!strcmp(m->name(),"full_upper")) {
      fullupper=*m;
      delete m;
      nnumcol=fullupper.nx();
      }
    else if (!strcmp(m->name(),"full_lower")) {
      fulllower=*m;
      delete m;
      }
    else {
      int k=index(m->row(),m->col());
      if (k<0||k>=nummirror) {
        fprintf(stderr,"PlaneMirror::PlaneMirror: index error: row=%d col=%d index=%d\n",
           m->row(),m->col(),k);
        Exit(1);
        }
      mirrors[k]=m;
      }
    }

  fullupper.xsize()=fullupper.xdiv()*fullupper.nx();
  fullupper.ysize()=fullupper.ydiv()*fullupper.ny();
  fulllower.xsize()=fulllower.xdiv()*fulllower.nx();
  fulllower.ysize()=fulllower.ydiv()*fulllower.ny();



  CPMI pm;
  pm.rec=radiuselectronhole;
  pm.iref=reflectivity().firstpoint();
  pm.nref=reflectivity().numpoints();
  ArteTable<CPMI>::push_back(pm);
  }
  

PlaneMirror::PlaneMirror(ArtePointer<CPMI> pm) {
  FullSimulation();

  radiuselectronhole=pm->rec;
  refl=Table(pm->iref,pm->nref);
  
  if (!(nummirror=ArteTable<CSPM>::nrow()-2)) return;
  
  fulllower=SinglePlaneMirror(ArtePointer<CSPM>(1));
  fullupper=SinglePlaneMirror(ArtePointer<CSPM>(2));
  fullupper.xsize()=fullupper.xdiv()*fullupper.nx();
  fullupper.ysize()=fullupper.ydiv()*fullupper.ny();
  fulllower.xsize()=fulllower.xdiv()*fulllower.nx();
  fulllower.ysize()=fulllower.ydiv()*fulllower.ny();

  nnumcol=fullupper.nx();

  mirrors=new SinglePlaneMirror*[nummirror];

  int i;
  SinglePlaneMirror* m;
  for (i=3;i<=nummirror+2;i++) {
    if (RICHdebug) { printf("plane mirror %d\n",i);fflush(stdout);}
    m=new SinglePlaneMirror(ArtePointer<CSPM>(i));
    if (RICHdebug) m->print();
    int k=index(m->row(),m->col());
    if (k<0||k>=nummirror) {
      fprintf(stderr,"PlaneMirror::PlaneMirror: index error: row=%d col=%d index=%d\n",
         m->row(),m->col(),k);
      Exit(1);
    }
    mirrors[k]=m;
  }  
}



PlaneMirror::~PlaneMirror() {
  int i;
  for (i=0;i<nummirror;i++) delete mirrors[i];
  delete mirrors;
  }

