#include <stdio.h>
#include <cmath>
#include <cstdio>
#include <vector>

#include <arte/arte.hh>
#include "silarte/silarte.hh"
#include "silarte/EVinfo.hh"

#include "grr.h"
#include "aframelib.h"
#include "sildat.h"
#include "holmes.h"

#include "Class_HOLMES.h"
#include "Class_TARGET.h"
#include "Class_GEDE.h"
#include "Struct_Track.h"
#include "Struct_Hitb.h"
#include "Struct_Gede.h"

//
// Fill Holmes' Tracks into RSEG table
// It's assumed that the first hit has lowest z and the last hit highest z
//
RSEG getRSEG( HOLMES_class* HOLMESp, int itrack )
{
  RSEG r;

  // get track
  Track_t *Track = HOLMESp->trackP( itrack, 0.);  
  if ( Track->nhit <= 0 )
    return r;  

  // start point at first hit
  Hitb_t* hitbf = HOLMESp->hitbP( Track->hit[0] );
  Track = HOLMESp->trackP( itrack, hitbf->gede->z );
  r.zf  = Track->z;
  r.xf  = Track->x;
  r.yf  = Track->y;
  r.txf = Track->tx;
  r.tyf = Track->ty;
  r.pf  = .1;
  for ( unsigned int i= 0; i<10; ++i ) r.cvf[i] = Track->C[i];
  for ( unsigned int i=10; i<14; ++i ) r.cvf[i] = 0;
  r.cvf[14] = 1e4;
 
  // end point at last hit
  Hitb_t* hitbe = HOLMESp->hitbP( Track->hit[Track->nhit-1] );
  Track = HOLMESp->trackP( itrack, hitbe->gede->z );
  r.ze  = Track->z;
  r.xe  = Track->x;
  r.ye  = Track->y;
  r.txe = Track->tx;
  r.tye = Track->ty;
  r.pe  = .1;
  for ( unsigned int i= 0; i<10; ++i ) r.cve[i] = Track->C[i];
  for ( unsigned int i=10; i<14; ++i ) r.cve[i] = 0;
  r.cve[14] = 1e4;

  // fill time and dtime with number of n-like and p-like sides
  // (distinguished by the pitch -> s.s. detectors are 'n-like'
  // THIS SHOULD ONLY BE A TEMPORARY FIX until there are better means
  // to access hit/layer information in RSEG !
  r.time = 0;
  r.dtim = 0;
  for ( int ihit=0; ihit<Track->nhit; ihit++ )
    if ( HOLMESp->hitbP( Track->hit[ihit] )->gede->pitch < 0.0053 )
      r.time++;      // n-like sides
    else
      r.dtim++;      // p-like sides

  // fill remaining quantities
  r.chi  = Track->chi2;
  r.fit  = 0;
  r.cmp  = 1;

  return r;
}


//
// Create RHITs for a segment
//
void RSEGtoRHIT( ArtePointer<RSEG> rseg )
{
  char *routine = "RSEGtoRHIT";
  vector<ArtePointer<RHIT> > rhits;

  if ( holmes_iol.debug_mode.value )
    aframe_mess(routine, GRR_SINFO, "entering RSEGtoRHIT...");

  for( ArteRelationIterator<HITB> hitb = rseg->hitb.begin();
       hitb != rseg->hitb.end(); ++hitb ) {

    RHIT rh;
    rh.x  = rseg->xf+((*hitb)->z-rseg->zf)*rseg->txf;
    rh.y  = rseg->yf+((*hitb)->z-rseg->zf)*rseg->tyf;
    rh.tx = rseg->txf;
    rh.ty = rseg->tyf;
    rh.p  = 0;
    rh.q  = 0;

    if ( holmes_iol.debug_mode.value ) {
      aframe_mess(routine, GRR_SINFO, "rh.x: %f rh.y: %f rh.tx: %f rh.ty: %f", rh.x, rh.y, rh.tx, rh.ty);
    }

    ArtePointer<RHIT> rhit = ArteTable<RHIT>::push_back( rh );
    if ( holmes_iol.debug_mode.value )
      aframe_mess(routine, GRR_SINFO, "Rhit: %d %d", rhit.index(), ArteTable<RHIT>::size()); 
    
    // createRelation( rseg,  rhit, &RSEG::rhit );
    rhits.push_back( rhit );
    createRelation( *hitb, rhit, &HITB::rhit );
  }

  for( vector<ArtePointer<RHIT> >::reverse_iterator rhit
	 = rhits.rbegin(); rhit!=rhits.rend(); ++rhit ) {
    if ( holmes_iol.debug_mode.value )
      aframe_mess(routine, GRR_SINFO, "create inverse relation");
    createRelation( rseg, *rhit, &RSEG::rhit );
  }
}

//
// =========================================
// silicon standalone reconstruction program 
// =========================================
//
int HolmesEvent()
{
  static HOLMES_class *HOLMESp = 0; 

  char *routine = "HolmesEvent";

  // Fill GEDE_class in case of MC
  if ( !HOLMESp && CheckMC() )
    sildat_notifyGede( HolmesFillGEDEfromARTE() );

  // initialize holmes
  HOLMESp = holmes_init();
  if ( !HOLMESp )
    return GRR_OK;

  // get target spot(s)
  float *xtarget, *ytarget, *ztarget;
  int   ntarget = HolmesGetTargets( &xtarget, &ytarget, &ztarget );

  double dx, dy;
  for ( int i=0; i<ntarget; i++ ) {
    // fill Target Information
    
    if(xtarget[i] == 0.) // target along x?
      dx = 0.35;
    else
      dx = 0.1;
    
    if(ytarget[i] == 0.) // target along y?
      dy = 0.35;
    else
      dy = 0.1;
    
    TARGET_class *trgP = new TARGET_class(xtarget[i], ytarget[i], ztarget[i],
					  dx, dy, 0.);
    HOLMESp->addTarget(i, trgP);
  }
  // Check if stored correctly:
  //  HOLMESp->printTargets();

  // fill Holmes-internal hit list
  HolmesFillHITBfromARTE( HOLMESp );

  // Do track finding
  HOLMESp->clearTracklist();
  if(ntarget == 1)
    HOLMESp->findTracks( xtarget[0], ytarget[0], ztarget[0], 1.);
  else
    HOLMESp->findTracks( 0., 0., 0., 1. );

  // print-out if requested
  if ( holmes_iol.debug_mode.value )
    aframe_mess(routine,GRR_SINFO, "Found %d hits and %d tracks.",
		HOLMESp->nhits(), HOLMESp->ntracks() );
  if ( holmes_iol.dump_hits.value )
    HOLMESp->dumpHitlist();
  if ( holmes_iol.dump_tracks.value )
    HOLMESp->dumpTracklist();

  // loop over tracks
  for ( int itrack=0; itrack<HOLMESp->ntracks(); itrack++ ) {

    // store track in RSEG table
    Track_t *Track = HOLMESp->trackP( itrack, 0 );

    ArteTable<RSEG>::push_back( getRSEG( HOLMESp, itrack ) );
    ArtePointer<RSEG> rseg( ArteTable<RSEG>::size() );

    // store corresponding hits in RHIT table
    for (int ihit=0; ihit<Track->nhit; ihit++) {

      int ihitb = HOLMESp->hitbP( Track->hit[ihit] )->hitb;

      ArtePointer<HITB> hitb( ihitb );
      createRelation( hitb, rseg, &RSEG::hitb );

    }

    // store track parameters at first point
    

    // update RHIT table 
    RSEGtoRHIT( rseg );

  }

  return GRR_OK;
}


// --------------------------------
//
//      Some Utility Routines
//
// --------------------------------

GEDE_class* HolmesFillGEDEfromARTE()
{
  double sxmn, sxmx, symn, symx, tmp, gcos, gsin;

  // Create empty GEDE class
  GEDE_class* GEDEp = new GEDE_class( (char*) 0 );

  // loop over GEDE planes
  for ( ArteTable<GEDE>::iterator igede = ArteTable<GEDE>::begin();
	igede != ArteTable<GEDE>::end(); igede++ ) {
    ArtePointer<GEDE> gede = *igede;
    if ( gede->cmp == 1 ) {
  
      // fill Gede_t structure with ARTE values
      Gede_t mygede;
      mygede.gede   = gede.index();
      mygede.side   = gede.index();
      char slname[5];
      strncpy( slname, gede->gesl->name, 4 ); 
      slname[4] = '\0';
      mygede.sl     = atoi( slname+2 );
      mygede.lay    = gede->lay;
      mygede.sec    = gede->sec;
      mygede.pitch  = gede->pit;

      // TG: Make Holmes compatible to all kind of geometries
      gcos = gede->cos  * gede->rot[0] - gede->sin * gede->rot[1];
      gsin = gede->sin  * gede->rot[4] - gede->cos * gede->rot[3];
      sxmn = gede->sxmn * gede->rot[0] +
             gede->symn * gede->rot[1];
      sxmx = gede->sxmx * gede->rot[0] +
             gede->symx * gede->rot[1];
      symn = gede->sxmn * gede->rot[3] +
             gede->symn * gede->rot[4];
      symx = gede->sxmx * gede->rot[3] +
             gede->symx * gede->rot[4];

      if( sxmn > sxmx ){
         tmp  = sxmx;
         sxmx = sxmn;
         sxmn = tmp;
      }
      if( symn > symx ){
         tmp  = symx;
         symx = symn;
         symn = tmp;
      }

      mygede.alpha  = 180. / M_PI * atan2( gsin, gcos );
      if ( mygede.alpha < -10 )
	mygede.alpha += 360.;
      mygede.s      = gsin;
      mygede.c      = gcos;
      mygede.d0     = - gede->cos * gede->sxmn
                      + gede->sin * gede->symn
	              + gede->xfw[0];
      mygede.dmin   = gede->xfw[0];
      mygede.sxmn   = sxmn;
      mygede.sxmx   = sxmx;
      mygede.symn   = symn;
      mygede.symx   = symx;

      // TG: End of comaptibility changes

      mygede.z      = ( gede->zmin + gede->zmax ) / 2.;      
      mygede.status = 1;            // all MC sides are perfect of course
      mygede.sign   = gede->xfw[2];
      *mygede.text  = '\0';
      
      // determine the type of view of the current plane
      if( -10 < mygede.alpha && mygede.alpha <  10 ||
	  170 < mygede.alpha && mygede.alpha < 190 ||
	  350 < mygede.alpha && mygede.alpha < 370   )
	mygede.view = 0;
      else
	mygede.view = 1;

      // Determine back side of the current detector
      //
      // !!!                                          !!!
      // !!!  here we use the assumption that layers  !!!
      // !!!  1 <-> 2 and 3 <-> 4 build pairs         !!!
      // !!!                                          !!!
      mygede.pair  = 0;
      for ( ArteTable<GEDE>::iterator igedeback = ArteTable<GEDE>::begin();
	    igedeback != ArteTable<GEDE>::end(); igedeback++ ) {
	ArtePointer<GEDE> gedeback = *igedeback;
	if (     gedeback->cmp == gede->cmp
	     && !strncmp( gedeback->gesl->name, gede->gesl->name, 4 )
	     &&  gedeback->sec == gede->sec
	     && (   gedeback->lay == 2 && gede->lay == 1
	         || gedeback->lay == 1 && gede->lay == 2
	         || gedeback->lay == 4 && gede->lay == 3
	         || gedeback->lay == 3 && gede->lay == 4 ) )
	  mygede.pair = gedeback.index();
      }

      // fill additional information of a gede struct
      mygede.dd    = 0;
      mygede.grad  = 0;   
      mygede.hesse = 0;   
      mygede.v[0]  =   mygede.c;
      mygede.v[1]  = - mygede.s;
      mygede.v[2]  =   mygede.c * mygede.z;
      mygede.v[3]  = - mygede.s * mygede.z;
      for (int i=0; i<4; i++)
	for (int j=0; j<4; j++)
	  mygede.vv[4*i+j] = mygede.v[i] * mygede.v[j];

      // now append to gede table
      GEDEp->gedeAppend( &mygede );
      
    }
  }

  return GEDEp;

}

//
//   Fill Holmes-internal hitlist from ARTE HITB table
//
void HolmesFillHITBfromARTE( HOLMES_class *HOLMESp )
{

  // clear the hit maps
  HOLMESp->clearHitlist();

  // refill the map from the HITB table
  for( ArteTable<HITB>::iterator iHITB = ArteTable<HITB>::begin();
       iHITB != ArteTable<HITB>::end(); iHITB++ ) {
    ArtePointer<HITB> ihit = *iHITB;

    if ( ihit->gede == 0 )              continue;

    Gede_t *GedeP = HOLMESp->gedeTableP( ihit->gede.index() );
    
    if ( !GedeP )                       continue;
    if ( ihit->q * GedeP->sign < 0 )    continue; 
    //    if ( fabs(ihit->q) > 100 )          continue;

    // fill next hitb into the map
    Hitb_t hitb;
    hitb.u      = ihit->x;
    hitb.cog    = -1;
    hitb.charge = ihit->q;
    hitb.sn     = ihit->y;
    hitb.flag   =  1;
    hitb.gede   = GedeP;
    hitb.width  =  1;
    hitb.hitb   = ihit.index();
    hitb.side   =  0;   // till we think of anything more clever

    HOLMESp->addtoHitlist(hitb);
  }

  return;
}
